import numpy as np
from collections import OrderedDict

from .base import labeller_func


@labeller_func(group_label='bounding_box')
def bounding_box_to_bounding_box(bbox):
    r"""
    Apply a single 'all' label to a given bounding box. This bounding
    box must be as specified by the :map:`bounding_box` method.
    """
    from menpo.shape import bounding_box

    mapping = OrderedDict()
    mapping['all'] = np.arange(4)
    return bounding_box(bbox.points[0], bbox.points[2]), mapping


@labeller_func(group_label='bounding_box')
def bounding_box_mirrored_to_bounding_box(bbox):
    r"""
    Apply a single 'all' label to a given bounding box that has been
    mirrored around the vertical axis (flipped around the Y-axis). This bounding
    box must be as specified by the :map:`bounding_box` method (but mirrored).
    """
    from menpo.shape import bounding_box

    mapping = OrderedDict()
    mapping['all'] = np.arange(4)
    return bounding_box(bbox.points[3], bbox.points[1]), mapping



import numpy as np
from numpy.testing import assert_allclose

from menpo.transform.thinplatesplines import ThinPlateSplines
from menpo.shape import PointCloud


square_src_landmarks = np.array([[-1.0, -1.0],
                                 [-1, 1],
                                 [1, -1],
                                 [1, 1]])

square_tgt_landmarks = np.array([[-1.0, -1.0],
                                 [-1, 1],
                                 [1, -1],
                                 [1, 1]])

perturbed_tgt_landmarks = np.array([[-0.6, -1.3],
                                    [-0.8, 1.2],
                                    [0.7, -0.8],
                                    [1.3, 0.5]])

x = np.arange(-1, 1, 0.01)
y = np.arange(-1, 1, 0.01)
xx, yy = np.meshgrid(x, y)
square_sample_points = np.array([xx.ravel(), yy.ravel()]).T

src = PointCloud(square_src_landmarks)
tgt = PointCloud(square_tgt_landmarks)
tgt_perturbed = PointCloud(perturbed_tgt_landmarks)


def test_tps_maps_src_to_tgt():
    tps = ThinPlateSplines(src, tgt_perturbed)
    assert_allclose(tps.apply(square_src_landmarks), perturbed_tgt_landmarks)


def test_tps_n_dims():
    tps = ThinPlateSplines(src, tgt_perturbed)
    assert(tps.n_dims == 2)


def test_tps_build_pseudoinverse():
    tps = ThinPlateSplines(src, tgt)
    tps_pinv = tps.pseudoinverse()

    assert (tps.source == tps_pinv.target)
    assert (tps.target == tps_pinv.source)
    # TODO: test the kernel


def test_tps_apply():
    src = PointCloud(np.array([[-1.0, -1.0], [-1, 1], [1, -1], [1, 1]]))
    tgt = PointCloud(np.array([[-2.0, -2.0], [-2, 2], [2, -2], [2, 2]]))
    pts = PointCloud(np.array([[-0.1, -1.0], [-0.5, 1.0], [2.1, -2.5]]))
    tps = ThinPlateSplines(src, tgt)
    result = tps.apply(pts)
    expected = np.array([[-0.2, -2.], [-1., 2.], [4.2, -5.]])
    assert_allclose(result.points, expected)


def test_tps_apply_batched():
    src = PointCloud(np.array([[-1.0, -1.0], [-1, 1], [1, -1], [1, 1]]))
    tgt = PointCloud(np.array([[-2.0, -2.0], [-2, 2], [2, -2], [2, 2]]))
    pts = PointCloud(np.array([[-0.1, -1.0], [-0.5, 1.0], [2.1, -2.5]]))
    tps = ThinPlateSplines(src, tgt)
    result = tps.apply(pts, batch_size=2)
    expected = np.array([[-0.2, -2.], [-1., 2.], [4.2, -5.]])
    assert_allclose(result.points, expected)



import numpy as np

from .base import HomogFamilyAlignment
from .affine import DiscreteAffine, Affine
from .similarity import Similarity


def Scale(scale_factor, n_dims=None):
    r"""
    Factory function for producing Scale transforms. Zero scale factors are not
    permitted.

    A :class:`UniformScale` will be produced if:

        - A `float` ``scale_factor`` and a ``n_dims`` `kwarg` are provided
        - A `ndarray` ``scale_factor`` with shape ``(n_dims,)`` is provided
          with all elements being the same

    A :class:`NonUniformScale` will be provided if:

        - A `ndarray` ``scale_factor`` with shape ``(n_dims,)`` is provided with
          at least two differing scale factors.

    Parameters
    ----------
    scale_factor : `float` or ``(n_dims,)`` `ndarray`
        Scale for each axis.
    n_dims : `int`, optional
        The dimensionality of the output transform.

    Returns
    -------
    scale : :class:`UniformScale` or :class:`NonUniformScale`
        The correct type of scale

    Raises
    ------
    ValueError
        If any of the scale factors is zero
    """
    from numbers import Number
    if not isinstance(scale_factor, Number):
        # some array like thing - make it a numpy array for sure
        scale_factor = np.asarray(scale_factor)
    if not np.all(scale_factor):
        raise ValueError('Having a zero in one of the scales is invalid')

    if n_dims is None:
        # scale_factor better be a numpy array then
        if np.allclose(scale_factor, scale_factor[0]):
            return UniformScale(scale_factor[0], scale_factor.shape[0])
        else:
            return NonUniformScale(scale_factor)
    else:
        # interpret as a scalar then
        return UniformScale(scale_factor, n_dims)


class NonUniformScale(DiscreteAffine, Affine):
    r"""
    An ``n_dims`` scale transform, with a scale component for each dimension.

    Parameters
    ----------
    scale : ``(n_dims,)`` `ndarray`
        A scale for each axis.
    skip_checks : `bool`, optional
        If ``True`` avoid sanity checks on ``h_matrix`` for performance.
    """

    def __init__(self, scale, skip_checks=False):
        scale = np.asarray(scale)
        if not skip_checks:
            if scale.size > 3 or scale.size < 2:
                raise ValueError("NonUniformScale can only be 2D or 3D"
                                 ", not {}".format(scale.size))
        h_matrix = np.eye(scale.size + 1)
        np.fill_diagonal(h_matrix, scale)
        h_matrix[-1, -1] = 1
        Affine.__init__(self, h_matrix, skip_checks=True, copy=False)

    @classmethod
    def init_identity(cls, n_dims):
        r"""
        Creates an identity transform.

        Parameters
        ----------
        n_dims : `int`
            The number of dimensions.

        Returns
        -------
        identity : :class:`NonUniformScale`
            The identity matrix transform.
        """
        return NonUniformScale(np.ones(n_dims))

    @property
    def scale(self):
        r"""
        The scale vector.

        :type: ``(n_dims,)`` `ndarray`
        """
        # Copy the vector as Numpy 1.10 will return a writeable view
        return self.h_matrix.diagonal()[:-1].copy()

    def _transform_str(self):
        message = 'NonUniformScale by {}'.format(self.scale)
        return message

    @property
    def n_parameters(self):
        """
        The number of parameters: ``n_dims``. They have the form
        ``[scale_x, scale_y, ....]`` representing the scale across each axis.

        :type: `list` of `int`
        """
        return self.scale.size

    def _as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as deltas from the identity warp. The parameters
        are output in the order ``[s0, s1, ...]``.

        +----------+--------------------------------------------+
        |parameter | definition                                 |
        +==========+============================================+
        |s0        | The scale across the first axis            |
        +----------+--------------------------------------------+
        |s1        | The scale across the second axis           |
        +----------+--------------------------------------------+
        |...       | ...                                        |
        +----------+--------------------------------------------+
        |sn        | The scale across the nth axis              |
        +----------+--------------------------------------------+

        Returns
        -------
        s : ``(n_dims,)`` `ndarray`
            The scale across each axis.
        """
        return self.scale

    def _from_vector_inplace(self, vector):
        r"""
        Updates the :class:`NonUniformScale` inplace.

        Parameters
        ----------
        vector : ``(n_dims,)`` `ndarray`
            The array of parameters.
        """
        np.fill_diagonal(self.h_matrix, vector)
        self.h_matrix[-1, -1] = 1

    @property
    def composes_inplace_with(self):
        r"""
        :class:`NonUniformScale` can swallow composition with any other
        :class:`NonUniformScale` and :class:`UniformScale`.
        """
        return NonUniformScale, UniformScale

    def pseudoinverse(self):
        """
        The inverse scale matrix.

        :type: :class:`NonUniformScale`
        """
        return NonUniformScale(1.0 / self.scale, skip_checks=True)


class UniformScale(DiscreteAffine, Similarity):
    r"""
    An abstract similarity scale transform, with a single scale component
    applied to all dimensions. This is abstracted out to remove unnecessary
    code duplication.

    Parameters
    ----------
    scale : ``(n_dims,)`` `ndarray`
        A scale for each axis.
    n_dims : `int`
        The number of dimensions
    skip_checks : `bool`, optional
        If ``True`` avoid sanity checks on ``h_matrix`` for performance.
    """

    def __init__(self, scale, n_dims, skip_checks=False):
        if not skip_checks:
            if n_dims > 3 or n_dims < 2:
                raise ValueError("UniformScale can only be 2D or 3D"
                                 ", not {}".format(n_dims))
        h_matrix = np.eye(n_dims + 1)
        np.fill_diagonal(h_matrix, scale)
        h_matrix[-1, -1] = 1
        Similarity.__init__(self, h_matrix, copy=False,
                            skip_checks=True)

    @classmethod
    def init_identity(cls, n_dims):
        r"""
        Creates an identity transform.

        Parameters
        ----------
        n_dims : `int`
            The number of dimensions.

        Returns
        -------
        identity : :class:`UniformScale`
            The identity matrix transform.
        """
        return UniformScale(1, n_dims)

    @property
    def scale(self):
        r"""
        The single scale value.

        :type: `float`
        """
        return self.h_matrix[0, 0]

    def _transform_str(self):
        message = 'UniformScale by {}'.format(self.scale)
        return message

    @property
    def n_parameters(self):
        r"""
        The number of parameters: 1

        :type: `int`
        """
        return 1

    def _as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as deltas from the identity warp. The parameters
        are output in the order ``[s]``.

        +----------+--------------------------------+
        |parameter | definition                     |
        +==========+================================+
        |s         | The scale across each axis     |
        +----------+--------------------------------+

        Returns
        -------
        s : `float`
            The scale across each axis.
        """
        return np.asarray(self.scale)

    def _from_vector_inplace(self, p):
        r"""
        Returns an instance of the transform from the given parameters,
        expected to be in Fortran ordering.

        Parameters
        ----------
        p : `float`
            The parameter
        """
        np.fill_diagonal(self.h_matrix, p)
        self.h_matrix[-1, -1] = 1

    @property
    def composes_inplace_with(self):
        r"""
        :class:`UniformScale` can swallow composition with any other
        :class:`UniformScale`.
        """
        return UniformScale

    def pseudoinverse(self):
        r"""
        The inverse scale.

        :type: :class:`UniformScale`
        """
        return UniformScale(1.0 / self.scale, self.n_dims, skip_checks=True)


class AlignmentUniformScale(HomogFamilyAlignment, UniformScale):
    r"""
    Constructs a :class:`UniformScale` by finding the optimal scale transform to
    align `source` to `target`.

    Parameters
    ----------
    source : :map:`PointCloud`
        The source pointcloud instance used in the alignment
    target : :map:`PointCloud`
        The target pointcloud instance used in the alignment
    """

    def __init__(self, source, target):
        HomogFamilyAlignment.__init__(self, source, target)
        UniformScale.__init__(self, target.norm() / source.norm(),
                              source.n_dims)

    def _from_vector_inplace(self, p):
        r"""
        Returns an instance of the transform from the given parameters,
        expected to be in Fortran ordering.

        Parameters
        ----------
        p : `float`
            The parameter
        """
        UniformScale._from_vector_inplace(self, p)
        self._sync_target_from_state()

    def _sync_state_from_target(self):
        new_scale = self.target.norm() / self.source.norm()
        np.fill_diagonal(self.h_matrix, new_scale)
        self.h_matrix[-1, -1] = 1

    def as_non_alignment(self):
        r"""Returns a copy of this uniform scale without it's alignment nature.

        Returns
        -------
        transform : :map:`UniformScale`
            A version of this scale with the same transform behavior but
            without the alignment logic.
        """
        return UniformScale(self.scale, self.n_dims)



from .face import (
    tongue_ibug_19_to_tongue_ibug_19,
    eye_ibug_close_17_to_eye_ibug_close_17,
    eye_ibug_close_17_to_eye_ibug_close_17_trimesh,
    eye_ibug_open_38_to_eye_ibug_open_38,
    eye_ibug_open_38_to_eye_ibug_open_38_trimesh,
    face_ibug_68_to_face_ibug_49, face_ibug_68_to_face_ibug_49_trimesh,
    face_ibug_68_to_face_ibug_51, face_ibug_68_to_face_ibug_51_trimesh,
    face_ibug_68_to_face_ibug_65,
    face_ibug_68_to_face_ibug_66, face_ibug_68_to_face_ibug_66_trimesh,
    face_ibug_68_to_face_ibug_68, face_ibug_68_to_face_ibug_68_trimesh,
    face_ibug_68_mirrored_to_face_ibug_68,
    face_ibug_49_to_face_ibug_49,
    face_imm_58_to_face_imm_58, face_lfpw_29_to_face_lfpw_29)
from .hand import hand_ibug_39_to_hand_ibug_39
from .face_3d import face_bu3dfe_83_to_face_bu3dfe_83
from .pose import (pose_flic_11_to_pose_flic_11,
                   pose_human36M_32_to_pose_human36M_17,
                   pose_human36M_32_to_pose_human36M_32,
                   pose_lsp_14_to_pose_lsp_14,
                   pose_stickmen_12_to_pose_stickmen_12)



from __future__ import division

import warnings

import numpy as np
from numpy.testing import assert_allclose
from pytest import raises, skip

from menpo.feature import (es, igo, daisy, no_op, normalize,
                           normalize_norm, normalize_std, normalize_var)
from menpo.image import Image, MaskedImage
from menpo.testing import is_same_array


def test_dsift_channels():
    try:
        from menpo.feature import dsift
    except ImportError:
        skip("Cyvlfeat must be installed to run this unit test")
    n_cases = 3
    num_bins_horizontal = np.random.randint(1, 3, [n_cases, 1])
    num_bins_vertical = np.random.randint(1, 3, [n_cases, 1])
    num_or_bins = np.random.randint(7, 9, [n_cases, 1])
    cell_size_horizontal = np.random.randint(1, 10, [n_cases, 1])
    cell_size_vertical = np.random.randint(1, 10, [n_cases, 1])
    channels = np.random.randint(1, 4, [n_cases])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(channels[i], 40, 40))
        dsift_img = dsift(image, window_step_horizontal=1,
                          window_step_vertical=1,
                          num_bins_horizontal=num_bins_horizontal[i, 0],
                          num_bins_vertical=num_bins_vertical[i, 0],
                          num_or_bins=num_or_bins[i, 0],
                          cell_size_horizontal=cell_size_horizontal[i, 0],
                          cell_size_vertical=cell_size_vertical[i, 0])
        n_channels = (num_bins_horizontal[i, 0] * num_bins_vertical[i, 0] *
                      num_or_bins[i, 0])
        assert_allclose(dsift_img.n_channels, n_channels)


def test_igo_channels():
    n_cases = 3
    channels = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(channels[i, 0], 40, 40))
        igo_img = igo(image)
        igo2_img = igo(image, double_angles=True)
        assert_allclose(igo_img.shape, image.shape)
        assert_allclose(igo2_img.shape, image.shape)
        assert_allclose(igo_img.n_channels, 2 * channels[i, 0])
        assert_allclose(igo2_img.n_channels, 4 * channels[i, 0])


def test_es_channels():
    n_cases = 3
    channels = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(channels[i, 0], 40, 40))
        es_img = es(image)
        assert_allclose(es_img.shape, image.shape)
        assert_allclose(es_img.n_channels, 2 * channels[i, 0])


def test_daisy_channels():
    n_cases = 3
    rings = np.random.randint(1, 3, [n_cases, 1])
    orientations = np.random.randint(1, 7, [n_cases, 1])
    histograms = np.random.randint(1, 6, [n_cases, 1])
    channels = np.random.randint(1, 5, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(channels[i, 0], 40, 40))
        daisy_img = daisy(image, step=4, rings=rings[i, 0],
                          orientations=orientations[i, 0],
                          histograms=histograms[i, 0])
        assert_allclose(daisy_img.shape, (3, 3))
        assert_allclose(daisy_img.n_channels,
                        ((rings[i, 0] * histograms[i, 0] + 1) * orientations[i, 0]))


def test_igo_values():
    image = Image([[1., 2.], [2., 1.]])
    igo_img = igo(image)
    res = np.array(
        [[[0.70710678, 0.70710678],
          [-0.70710678, -0.70710678]],
         [[0.70710678, -0.70710678],
          [0.70710678, -0.70710678]]])
    assert_allclose(igo_img.pixels, res)
    image = Image([[0., 0.], [0., 0.]])
    igo_img = igo(image)
    res = np.array([[[0., 0.], [0., 0.]], [[1., 1.], [1., 1.]]])
    assert_allclose(igo_img.pixels, res)


def test_es_values():
    image = Image([[1., 2.], [2., 1.]])
    es_img = es(image)
    k = 1.0 / (2 * (2 ** 0.5))
    res = np.array([[[k, -k], [k, -k]], [[k, k], [-k, -k]]])
    assert_allclose(es_img.pixels, res)


def test_daisy_values():
    image = Image([[1., 2., 3., 4.], [2., 1., 3., 4.], [1., 2., 3., 4.],
                   [2., 1., 3., 4.]])
    daisy_img = daisy(image, step=1, rings=2, radius=1, orientations=8,
                      histograms=8)
    assert_allclose(np.around(daisy_img.pixels[10, 0, 0], 6), 0.001355)
    assert_allclose(np.around(daisy_img.pixels[20, 0, 1], 6), 0.032237)
    assert_allclose(np.around(daisy_img.pixels[30, 1, 0], 6), 0.002032)
    assert_allclose(np.around(daisy_img.pixels[40, 1, 1], 6), 0.000163)


def test_dsift_values():
    try:
        from menpo.feature import dsift
    except ImportError:
        skip("Cyvlfeat must be installed to run this unit test")
        # Equivalent to the transpose of image in Matlab
    image = Image([[1, 2, 3, 4], [2, 1, 3, 4], [1, 2, 3, 4], [2, 1, 3, 4]])
    sift_img = dsift(image, cell_size_horizontal=2, cell_size_vertical=2)
    assert_allclose(np.around(sift_img.pixels[0, 0, 0], 6), 19.719786,
                    rtol=1e-04)
    assert_allclose(np.around(sift_img.pixels[1, 0, 1], 6), 141.535736,
                    rtol=1e-04)
    assert_allclose(np.around(sift_img.pixels[0, 1, 0], 6), 184.377472,
                    rtol=1e-04)
    assert_allclose(np.around(sift_img.pixels[5, 1, 1], 6), 39.04007,
                    rtol=1e-04)


def test_no_op():
    image = Image([[1., 2.], [2., 1.]])
    new_image = no_op(image)
    assert_allclose(image.pixels, new_image.pixels)
    assert not is_same_array(image.pixels, new_image.pixels)


def test_normalize_no_scale_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=None, mode='all')
    assert_allclose(new_image.pixels, pixels - 13.)


def test_normalize_norm_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_norm(image, mode='all')
    assert_allclose(np.linalg.norm(new_image.pixels), 1.)


def test_normalize_norm_channels():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_norm(image, mode='per_channel')
    assert_allclose(np.linalg.norm(new_image.pixels[0]), 1.)
    assert_allclose(np.linalg.norm(new_image.pixels[1]), 1.)
    assert_allclose(np.linalg.norm(new_image.pixels[2]), 1.)


def test_normalize_std_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_std(image, mode='all')
    assert_allclose(np.std(new_image.pixels), 1.)


def test_normalize_std_channels():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_std(image, mode='per_channel')
    assert_allclose(np.std(new_image.pixels[0]), 1.)
    assert_allclose(np.std(new_image.pixels[1]), 1.)
    assert_allclose(np.std(new_image.pixels[2]), 1.)


def test_normalize_var_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_var(image, mode='all')
    assert_allclose(np.var(new_image.pixels), 0.01648, atol=1e-3)


def test_normalize_var_channels():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_var(image, mode='per_channel')
    assert_allclose(np.var(new_image.pixels[0]), 0.15, atol=1e-5)
    assert_allclose(np.var(new_image.pixels[1]), 0.15, atol=1e-5)
    assert_allclose(np.var(new_image.pixels[2]), 0.15, atol=1e-5)


def test_normalize_no_scale_per_channel():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=None, mode='per_channel')
    assert_allclose(new_image.pixels[0], pixels[0] - 4.)
    assert_allclose(new_image.pixels[1], pixels[1] - 13.)
    assert_allclose(new_image.pixels[2], pixels[2] - 22.)


def test_normalize_no_scale_per_channel():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=None, mode='per_channel')
    assert_allclose(new_image.pixels[0], pixels[0] - 4.)
    assert_allclose(new_image.pixels[1], pixels[1] - 13.)
    assert_allclose(new_image.pixels[2], pixels[2] - 22.)


def test_normalize_scale_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    dummy_scale = lambda *a, **kwargs: np.array(2.0)
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=dummy_scale, mode='all')
    assert_allclose(new_image.pixels, (pixels - 13.0) / 2.0)


def test_normalize_scale_per_channel():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    dummy_scale = lambda *a, **kwargs: np.array(2.0)
    new_image = normalize(image, scale_func=dummy_scale, mode='per_channel')
    assert_allclose(new_image.pixels[0], (pixels[0] - 4.) / 2.0)
    assert_allclose(new_image.pixels[1], (pixels[1] - 13.) / 2.0)
    assert_allclose(new_image.pixels[2], (pixels[2] - 22.) / 2.0)


def test_normalize_unknown_mode_raises():
    image = Image.init_blank((2, 2))
    with raises(ValueError):
        normalize(image, mode='fake')


def test_normalize_0_variance_raises():
    image = Image.init_blank((2, 2))
    dummy_scale = lambda *a, **kwargs: np.array(0.0)
    with raises(ValueError):
        normalize(image, scale_func=dummy_scale)


def test_normalize_0_variance_warning():
    pixels = np.arange(8, dtype=np.float).reshape([2, 2, 2])
    image = Image(pixels, copy=False)
    dummy_scale = lambda *a, **kwargs: np.array([2.0, 0.0])
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        new_image = normalize(image, scale_func=dummy_scale,
                              error_on_divide_by_zero=False,
                              mode='per_channel')
    assert_allclose(new_image.pixels[0], [[-0.75, -0.25], [0.25, 0.75]])
    assert_allclose(new_image.pixels[1], [[-1.5, -0.5], [0.5, 1.5]])



from __future__ import division
import numpy as np
from scipy.ndimage import gaussian_filter


def _daisy(img, step=4, radius=15, rings=3, histograms=8, orientations=8,
           normalization='l1', sigmas=None, ring_radii=None):
    r"""Extract DAISY feature descriptors densely for the given image.

    DAISY is a feature descriptor similar to SIFT formulated in a way that
    allows for fast dense extraction. Typically, this is practical for
    bag-of-features image representations.

    The implementation follows Tola et al. [1]_ but deviate on the following
    points:

      * Histogram bin contribution are smoothed with a circular Gaussian
        window over the tonal range (the angular range).
      * The sigma values of the spatial Gaussian smoothing in this code do not
        match the sigma values in the original code by Tola et al. [2]_. In
        their code, spatial smoothing is applied to both the input image and
        the centre histogram. However, this smoothing is not documented in [1]_
        and, therefore, it is omitted.

    Parameters
    ----------
    img : (M, N) array
        Input image (greyscale).
    step : int, optional
        Distance between descriptor sampling points.
    radius : int, optional
        Radius (in pixels) of the outermost ring.
    rings : int, optional
        Number of rings.
    histograms  : int, optional
        Number of histograms sampled per ring.
    orientations : int, optional
        Number of orientations (bins) per histogram.
    normalization : [ 'l1' | 'l2' | 'daisy' | 'off' ], optional
        How to normalize the descriptors

          * 'l1': L1-normalization of each descriptor.
          * 'l2': L2-normalization of each descriptor.
          * 'daisy': L2-normalization of individual histograms.
          * 'off': Disable normalization.
    sigmas : 1D array of float, optional
        Standard deviation of spatial Gaussian smoothing for the centre
        histogram and for each ring of histograms. The array of sigmas should
        be sorted from the centre and out. I.e. the first sigma value defines
        the spatial smoothing of the centre histogram and the last sigma value
        defines the spatial smoothing of the outermost ring. Specifying sigmas
        overrides the following parameter.

            ``rings = len(sigmas) - 1``
    ring_radii : 1D array of int, optional
        Radius (in pixels) for each ring. Specifying ring_radii overrides the
        following two parameters.

            ``rings = len(ring_radii)``
            ``radius = ring_radii[-1]``

        If both sigmas and ring_radii are given, they must satisfy the
        following predicate since no radius is needed for the centre
        histogram.

            ``len(ring_radii) == len(sigmas) + 1``

    Returns
    -------
    descs : array
        Grid of DAISY descriptors for the given image as an array
        dimensionality  (P, Q, R) where ::

            ``P = ceil((M - radius*2) / step)``
            ``Q = ceil((N - radius*2) / step)``
            ``R = (rings * histograms + 1) * orientations``

    descs_img : (M, N, 3) array (only if visualize==True)
        Visualization of the DAISY descriptors.

    References
    ----------
    .. [1] Tola et al. "Daisy: An efficient dense descriptor applied to wide-
           baseline stereo." Pattern Analysis and Machine Intelligence, IEEE
           Transactions on 32.5 (2010): 815-830.
    .. [2] http://cvlab.epfl.ch/alumni/tola/daisy.html
    """
    from menpo.feature import gradient

    # Compute image derivatives.
    # Get number of input image's channels
    n_channels = img.shape[0]

    # Compute image gradient
    grad = gradient(img)

    # For each pixel, select gradient with highest magnitude
    grad_mag = np.zeros(img.shape[1:])
    grad_ori = np.zeros(img.shape[1:])
    for c in range(n_channels):
        c_grad_mag = np.sqrt(grad[c] ** 2 + grad[c + n_channels] ** 2)
        tmp_max_mask = c_grad_mag > grad_mag
        grad_mag[tmp_max_mask] = c_grad_mag[tmp_max_mask]
        grad_ori[tmp_max_mask] = np.arctan2(grad[c][tmp_max_mask],
                                            grad[c + n_channels][tmp_max_mask])

    orientation_kappa = orientations / np.pi
    orientation_angles = [2 * o * np.pi / orientations - np.pi
                          for o in range(orientations)]
    hist = np.empty((orientations,) + img.shape[1:], dtype=float)
    for i, o in enumerate(orientation_angles):
        # Weigh bin contribution by the circular normal distribution
        hist[i] = np.exp(orientation_kappa * np.cos(grad_ori - o))
        # Weigh bin contribution by the gradient magnitude
        hist[i] = hist[i] * grad_mag

    # Smooth orientation histograms for the center and all rings.
    sigmas = [sigmas[0]] + sigmas
    hist_smooth = np.empty((rings + 1,) + hist.shape, dtype=float)
    for i in range(rings + 1):
        for j in range(orientations):
            hist_smooth[i, j] = gaussian_filter(hist[j], sigma=sigmas[i])

    # Assemble descriptor grid.
    theta = [2 * np.pi * j / histograms for j in range(histograms)]
    desc_dims = (rings * histograms + 1) * orientations
    descs = np.empty((desc_dims, img.shape[1] - 2 * radius,
                      img.shape[2] - 2 * radius))
    descs[:orientations] = hist_smooth[0, :, radius:-radius, radius:-radius]
    idx = orientations
    for i in range(rings):
        for j in range(histograms):
            y_min = radius + int(np.round(ring_radii[i] * np.sin(theta[j])))
            y_max = descs.shape[1] + y_min
            x_min = radius + int(np.round(ring_radii[i] * np.cos(theta[j])))
            x_max = descs.shape[2] + x_min
            descs[idx:idx + orientations] = hist_smooth[i + 1, :,
                                                        y_min:y_max,
                                                        x_min:x_max]
            idx += orientations
    descs = descs[:, ::step, ::step]

    # Normalize descriptors.
    if normalization != 'off':
        descs += 1e-10
        if normalization == 'l1':
            descs /= np.sum(descs, axis=0)
        elif normalization == 'l2':
            descs /= np.sqrt(np.sum(descs ** 2, axis=0))
        elif normalization == 'daisy':
            for i in range(0, desc_dims, orientations):
                norms = np.sqrt(np.sum(descs[i:i + orientations] ** 2, axis=0))
                descs[i:i + orientations] /= norms

    descs = np.require(descs, requirements=['C'])

    return descs



import numpy as np

from .base import ndfeature


@ndfeature
def sum_channels(pixels, channels=None):
    r"""
    Create the sum of the channels of an image that can be used for
    visualization.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels.
    channels : `list` of `int` or ``None``
        The list of channels to be used. If ``None``, then all the channels are
        employed.
    """
    # if channels is None, then all channels are used
    if channels is None:
        # Not indexing is twice as fast
        sum_image = np.sum(pixels, axis=0)
    else:
        sum_image = np.sum(pixels[channels], axis=0)
    return sum_image.reshape((1,) + sum_image.shape)  # add a channel axis



from __future__ import division

import itertools
import warnings

import numpy as np

scipy_gaussian_filter = None  # expensive

from .base import ndfeature, imgfeature


@ndfeature
def gradient(pixels):
    r"""
    Calculates the gradient of an input image. The image is assumed to have
    channel information on the first axis. In the case of multiple channels,
    it returns the gradient over each axis over each channel as the first axis.

    The gradient is computed using second order accurate central differences in
    the interior and first order accurate one-side (forward or backwards)
    differences at the boundaries.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array where the first dimension
        is interpreted as channels. This means an N-dimensional image is
        represented by an N+1 dimensional array.
        If the image is 2-dimensional the pixels should be of type
        float/double (int is not supported).

    Returns
    -------
    gradient : `ndarray`
        The gradient over each axis over each channel. Therefore, the
        first axis of the gradient of a 2D, single channel image, will have
        length `2`. The first axis of the gradient of a 2D, 3-channel image,
        will have length `6`, the ordering being
        ``I[:, 0, 0] = [R0_y, G0_y, B0_y, R0_x, G0_x, B0_x]``. To be clear,
        all the ``y``-gradients are returned over each channel, then all
        the ``x``-gradients.
    """
    if pixels.dtype == np.uint8:
        raise TypeError("Attempting to take the gradient on a uint8 image.")
    n_dims = pixels.ndim - 1
    grad_per_dim_per_channel = [np.gradient(g, edge_order=1)
                                for g in pixels]
    # Flatten out the separate dims
    grad_per_channel = list(itertools.chain.from_iterable(
        grad_per_dim_per_channel))
    # Add a channel axis for broadcasting
    grad_per_channel = [g[None, ...] for g in grad_per_channel]

    # Permute the list so it is first axis, second axis, etc
    grad_per_channel = [grad_per_channel[i::n_dims]
                        for i in range(n_dims)]
    grad_per_channel = list(itertools.chain.from_iterable(grad_per_channel))

    # Concatenate gradient list into an array (the new_image)
    return np.concatenate(grad_per_channel, axis=0)


@ndfeature
def gaussian_filter(pixels, sigma):
    r"""
    Calculates the convolution of the input image with a multidimensional
    Gaussian filter.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    sigma : `float` or `list` of `float`
        The standard deviation for Gaussian kernel. The standard deviations of
        the Gaussian filter are given for each axis as a `list`, or as a single
        `float`, in which case it is equal for all axes.

    Returns
    -------
    output_image : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        The filtered image has the same type and size as the input ``pixels``.
    """
    global scipy_gaussian_filter
    if scipy_gaussian_filter is None:
        from scipy.ndimage import gaussian_filter as scipy_gaussian_filter
    output = np.empty(pixels.shape, dtype=pixels.dtype)
    for dim in range(pixels.shape[0]):
        scipy_gaussian_filter(pixels[dim], sigma, output=output[dim])
    return output


@ndfeature
def igo(pixels, double_angles=False, verbose=False):
    r"""
    Extracts Image Gradient Orientation (IGO) features from the input image.
    The output image has ``N * C`` number of channels, where ``N`` is the
    number of channels of the original image and ``C = 2`` or ``C = 4``
    depending on whether double angles are used.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    double_angles : `bool`, optional
        Assume that ``phi`` represents the gradient orientations.

        If this flag is ``False``, the features image is the concatenation of
        ``cos(phi)`` and ``sin(phi)``, thus 2 channels.

        If ``True``, the features image is the concatenation of
        ``cos(phi)``, ``sin(phi)``, ``cos(2 * phi)``, ``sin(2 * phi)``, thus 4
        channels.
    verbose : `bool`, optional
        Flag to print IGO related information.

    Returns
    -------
    igo : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        The IGO features image. It has the same type and shape as the input
        ``pixels``. The output number of channels depends on the
        ``double_angles`` flag.

    Raises
    ------
    ValueError
        Image has to be 2D in order to extract IGOs.

    References
    ----------
    .. [1] G. Tzimiropoulos, S. Zafeiriou and M. Pantic, "Subspace learning
        from image gradient orientations", IEEE Transactions on Pattern Analysis
        and Machine Intelligence, vol. 34, num. 12, p. 2454--2466, 2012.
    """
    # check number of dimensions
    if len(pixels.shape) != 3:
        raise ValueError('IGOs only work on 2D images. Expects image data '
                         'to be 3D, channels + shape.')
    n_img_chnls = pixels.shape[0]
    # feature channels per image channel
    feat_chnls = 2
    if double_angles:
        feat_chnls = 4

    # compute gradients
    grad = gradient(pixels)
    # compute angles
    grad_orient = np.angle(grad[:n_img_chnls] + 1j * grad[n_img_chnls:])
    # compute igo image
    igo_pixels = np.empty((n_img_chnls * feat_chnls,
                           pixels.shape[1], pixels.shape[2]),
                          dtype=pixels.dtype)

    if double_angles:
        dbl_grad_orient = 2 * grad_orient
        # y angles
        igo_pixels[:n_img_chnls] = np.sin(grad_orient)
        igo_pixels[n_img_chnls:n_img_chnls * 2] = np.sin(dbl_grad_orient)

        # x angles
        igo_pixels[n_img_chnls * 2:n_img_chnls * 3] = np.cos(grad_orient)
        igo_pixels[n_img_chnls * 3:] = np.cos(dbl_grad_orient)
    else:
        igo_pixels[:n_img_chnls] = np.sin(grad_orient)  # y
        igo_pixels[n_img_chnls:] = np.cos(grad_orient)  # x

    # print information
    if verbose:
        info_str = "IGO Features:\n"
        info_str = "{}  - Input image is {}W x {}H with {} channels.\n".format(
            info_str, pixels.shape[2], pixels.shape[1], n_img_chnls)
        info_str = "{}  - Double angles are {}.\n".format(
            info_str, 'enabled' if double_angles else 'disabled')
        info_str = "{}Output image size {}W x {}H with {} channels.".format(
            info_str, igo_pixels.shape[2], igo_pixels.shape[1], n_img_chnls)
        print(info_str)
    return igo_pixels


@ndfeature
def es(pixels, verbose=False):
    r"""
    Extracts Edge Structure (ES) features from the input image. The output image
    has ``N * C`` number of channels, where ``N`` is the number of channels of
    the original image and ``C = 2``.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either an image object itself or an array where the first axis
        represents the number of channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    verbose : `bool`, optional
        Flag to print ES related information.

    Returns
    -------
    es : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        The ES features image. It has the same type and shape as the input
        ``pixels``. The output number of channels is ``C = 2``.

    Raises
    ------
    ValueError
        Image has to be 2D in order to extract ES features.

    References
    ----------
    .. [1] T. Cootes, C. Taylor, "On representing edge structure for model
        matching", Proceedings of the IEEE Conference on Computer Vision and
        Pattern Recognition (CVPR), 2001.
    """
    # check number of dimensions
    if len(pixels.shape) != 3:
        raise ValueError('ES features only work on 2D images. Expects '
                         'image data to be 3D, channels + shape.')
    n_img_chnls = pixels.shape[0]
    # feature channels per image channel
    feat_channels = 2
    # compute gradients
    grad = gradient(pixels)
    # compute magnitude
    grad_abs = np.abs(grad[:n_img_chnls] + 1j * grad[n_img_chnls:])
    # compute es image
    grad_abs = grad_abs + np.median(grad_abs)
    es_pixels = np.empty((pixels.shape[0] * feat_channels,
                          pixels.shape[1], pixels.shape[2]),
                         dtype=pixels.dtype)

    es_pixels[:n_img_chnls] = grad[:n_img_chnls] / grad_abs
    es_pixels[n_img_chnls:] = grad[n_img_chnls:] / grad_abs

    # print information
    if verbose:
        info_str = "ES Features:\n"
        info_str = "{}  - Input image is {}W x {}H with {} channels.\n".format(
            info_str, pixels.shape[2], pixels.shape[1], n_img_chnls)
        info_str = "{}Output image size {}W x {}H with {} channels.".format(
            info_str, es_pixels.shape[2], es_pixels.shape[1], n_img_chnls)
        print(info_str)
    return es_pixels


@ndfeature
def daisy(pixels, step=1, radius=15, rings=2, histograms=2, orientations=8,
          normalization='l1', sigmas=None, ring_radii=None, verbose=False):
    r"""
    Extracts Daisy features from the input image. The output image has ``N * C``
    number of channels, where ``N`` is the number of channels of the original
    image and ``C`` is the feature channels determined by the input options.
    Specifically, ``C = (rings * histograms + 1) * orientations``.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    step : `int`, optional
        The sampling step that defines the density of the output image.
    radius : `int`, optional
        The radius (in pixels) of the outermost ring.
    rings : `int`, optional
        The number of rings to be used.
    histograms : `int`, optional
        The number of histograms sampled per ring.
    orientations : `int`, optional
        The number of orientations (bins) per histogram.
    normalization : [ 'l1', 'l2', 'daisy', None ], optional
        It defines how to normalize the descriptors
        If 'l1' then L1-normalization is applied at each descriptor.
        If 'l2' then L2-normalization is applied at each descriptor.
        If 'daisy' then L2-normalization is applied at individual histograms.
        If None then no normalization is employed.
    sigmas : `list` of `float` or ``None``, optional
        Standard deviation of spatial Gaussian smoothing for the centre
        histogram and for each ring of histograms. The `list` of sigmas should
        be sorted from the centre and out. I.e. the first sigma value defines
        the spatial smoothing of the centre histogram and the last sigma value
        defines the spatial smoothing of the outermost ring. Specifying sigmas
        overrides the `rings` parameter by setting ``rings = len(sigmas) - 1``.
    ring_radii : `list` of `float` or ``None``, optional
        Radius (in pixels) for each ring. Specifying `ring_radii` overrides the
        `rings` and `radius` parameters by setting ``rings = len(ring_radii)``
        and ``radius = ring_radii[-1]``.

        If both sigmas and ring_radii are given, they must satisfy ::

            len(ring_radii) == len(sigmas) + 1

        since no radius is needed for the centre histogram.
    verbose : `bool`
        Flag to print Daisy related information.

    Returns
    -------
    daisy : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        The ES features image. It has the same type and shape as the input
        ``pixels``. The output number of channels is
        ``C = (rings * histograms + 1) * orientations``.

    Raises
    ------
    ValueError
        len(sigmas)-1 != len(ring_radii)
    ValueError
        Invalid normalization method.

    References
    ----------
    .. [1] E. Tola, V. Lepetit and P. Fua, "Daisy: An efficient dense descriptor
        applied to wide-baseline stereo", IEEE Transactions on Pattern Analysis
        and Machine Intelligence, vol. 32, num. 5, p. 815-830, 2010.
    """
    from menpo.external.skimage._daisy import _daisy

    # Parse options
    if sigmas is not None and ring_radii is not None \
            and len(sigmas) - 1 != len(ring_radii):
        raise ValueError('`len(sigmas)-1 != len(ring_radii)`')
    if ring_radii is not None:
        rings = len(ring_radii)
        radius = ring_radii[-1]
    if sigmas is not None:
        rings = len(sigmas) - 1
    if sigmas is None:
        sigmas = [radius * (i + 1) / float(2 * rings) for i in range(rings)]
    if ring_radii is None:
        ring_radii = [radius * (i + 1) / float(rings) for i in range(rings)]
    if normalization is None:
        normalization = 'off'
    if normalization not in ['l1', 'l2', 'daisy', 'off']:
        raise ValueError('Invalid normalization method.')

    # Compute daisy features
    daisy_descriptor = _daisy(pixels, step=step, radius=radius, rings=rings,
                              histograms=histograms, orientations=orientations,
                              normalization=normalization, sigmas=sigmas,
                              ring_radii=ring_radii)

    # print information
    if verbose:
        info_str = "Daisy Features:\n"
        info_str = "{}  - Input image is {}W x {}H with {} channels.\n".format(
            info_str, pixels.shape[2], pixels.shape[1], pixels.shape[0])
        info_str = "{}  - Sampling step is {}.\n".format(info_str, step)
        info_str = "{}  - Radius of {} pixels, {} rings and {} histograms " \
                   "with {} orientations.\n".format(
            info_str, radius, rings, histograms, orientations)
        if not normalization == 'off':
            info_str = "{}  - Using {} normalization.\n".format(info_str,
                                                                normalization)
        else:
            info_str = "{}  - No normalization emplyed.\n".format(info_str)
        info_str = "{}Output image size {}W x {}H x {}.".format(
            info_str, daisy_descriptor.shape[2], daisy_descriptor.shape[1],
            daisy_descriptor.shape[0])
        print(info_str)

    return daisy_descriptor


@imgfeature
def normalize(img, scale_func=None, mode='all',
              error_on_divide_by_zero=True):
    r"""
    Normalize the pixel values via mean centering and an optional scaling. By
    default the scaling will be ``1.0``. The ``mode`` parameter selects
    whether the normalisation is computed across all pixels in the image or
    per-channel.

    Parameters
    ----------
    img : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    scale_func : `callable`, optional
        Compute the scaling factor. Expects a single parameter and an optional
        `axis` keyword argument and will be passed the entire pixel array.
        Should return a 1D numpy array of one or more values.
    mode : ``{all, per_channel}``, optional
        If ``all``, the normalization is over all channels. If
        ``per_channel``, each channel individually is mean centred and
        normalized in variance.
    error_on_divide_by_zero : `bool`, optional
        If ``True``, will raise a ``ValueError`` on dividing by zero.
        If ``False``, will merely raise a warning and only those values
        with non-zero denominators will be normalized.

    Returns
    -------
    pixels : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        A normalized copy of the image that was passed in.

    Raises
    ------
    ValueError
        If any of the denominators are 0 and ``error_on_divide_by_zero`` is
        ``True``.
    """
    if scale_func is None:
        def scale_func(_, axis=None):
            return np.array([1.0])

    pixels = img.as_vector(keep_channels=True)

    if mode == 'all':
        centered_pixels = pixels - np.mean(pixels)
        scale_factor = scale_func(centered_pixels)
    elif mode == 'per_channel':
        centered_pixels = pixels - np.mean(pixels, axis=1, keepdims=True)
        scale_factor = scale_func(centered_pixels, axis=1).reshape([-1, 1])
    else:
        raise ValueError("Supported modes are {{'all', 'per_channel'}} - '{}' "
                         "is not known".format(mode))

    zero_denom = (scale_factor == 0).ravel()
    any_non_zero = np.any(zero_denom)
    if error_on_divide_by_zero and any_non_zero:
        raise ValueError("Computed scale factor cannot be 0.0")
    elif any_non_zero:
        warnings.warn('One or more the scale factors are 0.0 and thus these'
                      'entries will be skipped during normalization.')
        non_zero_denom = ~zero_denom
        centered_pixels[non_zero_denom] = (centered_pixels[non_zero_denom] /
                                           scale_factor[non_zero_denom])
        return img.from_vector(centered_pixels)
    else:
        return img.from_vector(centered_pixels / scale_factor)


@ndfeature
def normalize_norm(pixels, mode='all', error_on_divide_by_zero=True):
    r"""
    Normalize the pixels to be mean centred and have unit norm. The ``mode``
    parameter selects whether the normalisation is computed across all pixels in
    the image or per-channel.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    mode : ``{all, per_channel}``, optional
        If ``all``, the normalization is over all channels. If
        ``per_channel``, each channel individually is mean centred and
        normalized in variance.
    error_on_divide_by_zero : `bool`, optional
        If ``True``, will raise a ``ValueError`` on dividing by zero.
        If ``False``, will merely raise a warning and only those values
        with non-zero denominators will be normalized.

    Returns
    -------
    pixels : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        A normalized copy of the image that was passed in.

    Raises
    ------
    ValueError
        If any of the denominators are 0 and ``error_on_divide_by_zero`` is
        ``True``.
    """

    def unit_norm(x, axis=None):
        return np.linalg.norm(x, axis=axis)

    return normalize(pixels, scale_func=unit_norm, mode=mode,
                     error_on_divide_by_zero=error_on_divide_by_zero)


@ndfeature
def normalize_std(pixels, mode='all', error_on_divide_by_zero=True):
    r"""
    Normalize the pixels to be mean centred and have unit standard deviation.
    The ``mode`` parameter selects whether the normalisation is computed across
    all pixels in the image or per-channel.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    mode : ``{all, per_channel}``, optional
        If ``all``, the normalization is over all channels. If
        ``per_channel``, each channel individually is mean centred and
        normalized in variance.
    error_on_divide_by_zero : `bool`, optional
        If ``True``, will raise a ``ValueError`` on dividing by zero.
        If ``False``, will merely raise a warning and only those values
        with non-zero denominators will be normalized.

    Returns
    -------
    pixels : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        A normalized copy of the image that was passed in.

    Raises
    ------
    ValueError
        If any of the denominators are 0 and ``error_on_divide_by_zero`` is
        ``True``.
    """

    def unit_std(x, axis=None):
        return np.std(x, axis=axis)

    return normalize(pixels, scale_func=unit_std, mode=mode,
                     error_on_divide_by_zero=error_on_divide_by_zero)


@ndfeature
def normalize_var(pixels, mode='all', error_on_divide_by_zero=True):
    r"""
    Normalize the pixels to be mean centred and normalize according
    to the variance.
    The ``mode`` parameter selects whether the normalisation is computed across
    all pixels in the image or per-channel.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.
    mode : ``{all, per_channel}``, optional
        If ``all``, the normalization is over all channels. If
        ``per_channel``, each channel individually is mean centred and
        normalized in variance.
    error_on_divide_by_zero : `bool`, optional
        If ``True``, will raise a ``ValueError`` on dividing by zero.
        If ``False``, will merely raise a warning and only those values
        with non-zero denominators will be normalized.

    Returns
    -------
    pixels : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        A normalized copy of the image that was passed in.

    Raises
    ------
    ValueError
        If any of the denominators are 0 and ``error_on_divide_by_zero`` is
        ``True``.
    """

    def unit_var(x, axis=None):
        return np.var(x, axis=axis)

    return normalize(pixels, scale_func=unit_var, mode=mode,
                     error_on_divide_by_zero=error_on_divide_by_zero)


@ndfeature
def no_op(pixels):
    r"""
    A no operation feature - does nothing but return a copy of the pixels
    passed in.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels. This means an N-dimensional image
        is represented by an N+1 dimensional array.

    Returns
    -------
    pixels : :map:`Image` or subclass or ``(X, Y, ..., Z, C)`` `ndarray`
        A copy of the image that was passed in.
    """
    return pixels.copy()


def features_selection_widget():
    r"""
    Widget that allows for easy selection of a features function and its
    options. It also has a 'preview' tab for visual inspection. It returns a
    `list` of length 1 with the selected features function closure.

    Returns
    -------
    features_function : `list` of length ``1``
        The function closure of the features function using `functools.partial`.
        So the function can be called as: ::

            features_image = features_function[0](image)

    Examples
    --------
    The widget can be invoked as ::

        from menpo.feature import features_selection_widget
        features_fun = features_selection_widget()

    And the returned function can be used as ::

        import menpo.io as mio
        image = mio.import_builtin_asset.lenna_png()
        features_image = features_fun[0](image)
    """
    from menpowidgets import features_selection

    return features_selection()



# Parts of this code taken from:
#
# Copyright (c) 2006-2015, Christoph Gohlke
# Copyright (c) 2006-2015, The Regents of the University of California
# Produced at the Laboratory for Fluorescence Dynamics
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import numpy as np

from .base import HomogFamilyAlignment
from .affine import DiscreteAffine
from .similarity import Similarity


def optimal_rotation_matrix(source, target, allow_mirror=False):
    r"""
    Performs an SVD on the correlation matrix to find an optimal rotation
    between `source` and `target`.

    Parameters
    ----------
    source: :map:`PointCloud`
        The source points to be aligned
    target: :map:`PointCloud`
        The target points to be aligned
    allow_mirror : `bool`, optional
        If ``True``, the Kabsch algorithm check is not performed, and mirroring
        of the Rotation matrix is permitted.

    Returns
    -------
    rotation : `ndarray`
        The optimal square rotation matrix.
    """
    correlation = np.dot(target.points.T, source.points)
    U, D, Vt = np.linalg.svd(correlation)
    R = np.dot(U, Vt)

    if not allow_mirror:
        # d = sgn(det(V * Ut))
        d = np.sign(np.linalg.det(R))
        if d < 0:
            E = np.eye(U.shape[0])
            E[-1, -1] = d
            # R = U * E * Vt, E = [[1, 0, 0], [0, 1, 0], [0, 0, d]] for 2D
            R = np.dot(U, np.dot(E, Vt))
    return R


# TODO build rotations about axis, euler angles etc
# see http://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
# for details

class Rotation(DiscreteAffine, Similarity):
    r"""
    Abstract `n_dims` rotation transform.

    Parameters
    ----------
    rotation_matrix : ``(n_dims, n_dims)`` `ndarray`
        A valid, square rotation matrix
    skip_checks : `bool`, optional
        If ``True`` avoid sanity checks on ``rotation_matrix`` for performance.
    """

    def __init__(self, rotation_matrix, skip_checks=False):
        h_matrix = np.eye(rotation_matrix.shape[0] + 1)
        Similarity.__init__(self, h_matrix, copy=False, skip_checks=True)
        self.set_rotation_matrix(rotation_matrix, skip_checks=skip_checks)

    @classmethod
    def init_identity(cls, n_dims):
        r"""
        Creates an identity transform.

        Parameters
        ----------
        n_dims : `int`
            The number of dimensions.

        Returns
        -------
        identity : :class:`Rotation`
            The identity matrix transform.
        """
        return Rotation(np.eye(n_dims))

    @classmethod
    def init_from_2d_ccw_angle(cls, theta, degrees=True):
        r"""
        Convenience constructor for 2D CCW rotations about the origin.

        Parameters
        ----------
        theta : `float`
            The angle of rotation about the origin
        degrees : `bool`, optional
            If ``True`` theta is interpreted as a degree. If ``False``, theta is
            interpreted as radians.

        Returns
        -------
        rotation : :map:`Rotation`
            A 2D rotation transform.
        """
        if degrees:
            theta = np.deg2rad(theta)
        return Rotation(np.array([[np.cos(theta), -np.sin(theta)],
                                  [np.sin(theta),  np.cos(theta)]]),
                        skip_checks=True)

    @classmethod
    def init_3d_from_quaternion(cls, q):
        r"""
        Convenience constructor for 3D rotations based on quaternion parameters.

        Parameters
        ----------
        q : ``(4,)`` `ndarray`
            The quaternion parameters.

        Returns
        -------
        rotation : :map:`Rotation`
            A 3D rotation transform.
        """
        r = cls.init_identity(n_dims=3)
        return r.from_vector(q)

    @classmethod
    def init_from_3d_ccw_angle_around_x(cls, theta, degrees=True):
        r"""
        Convenience constructor for 3D CCW rotations around the x axis

        Parameters
        ----------
        theta : `float`
            The angle of rotation about the origin
        degrees : `bool`, optional
            If ``True`` theta is interpreted as a degree. If ``False``, theta is
            interpreted as radians.

        Returns
        -------
        rotation : :map:`Rotation`
            A 3D rotation transform.
        """
        if degrees:
            theta = np.deg2rad(theta)
        return Rotation(np.array([[ 1,             0,              0],
                                  [ 0, np.cos(theta), -np.sin(theta)],
                                  [ 0, np.sin(theta), np.cos(theta)]]),
                        skip_checks=True)

    @classmethod
    def init_from_3d_ccw_angle_around_y(cls, theta, degrees=True):
        r"""
        Convenience constructor for 3D CCW rotations around the y axis

        Parameters
        ----------
        theta : `float`
            The angle of rotation about the origin
        degrees : `bool`, optional
            If ``True`` theta is interpreted as a degree. If ``False``, theta is
            interpreted as radians.

        Returns
        -------
        rotation : :map:`Rotation`
            A 3D rotation transform.
        """
        if degrees:
            theta = np.deg2rad(theta)
        return Rotation(np.array([[ np.cos(theta), 0, np.sin(theta)],
                                  [             0, 1,             0],
                                  [-np.sin(theta), 0, np.cos(theta)]]),
                        skip_checks=True)

    @classmethod
    def init_from_3d_ccw_angle_around_z(cls, theta, degrees=True):
        r"""
        Convenience constructor for 3D CCW rotations around the z axis

        Parameters
        ----------
        theta : `float`
            The angle of rotation about the origin
        degrees : `bool`, optional
            If ``True`` theta is interpreted as a degree. If ``False``, theta is
            interpreted as radians.

        Returns
        -------
        rotation : :map:`Rotation`
            A 3D rotation transform.
        """
        if degrees:
            theta = np.deg2rad(theta)
        return Rotation(np.array([[ np.cos(theta), -np.sin(theta), 0],
                                  [ np.sin(theta),  np.cos(theta), 0],
                                  [             0,              0, 1]]),
                        skip_checks=True)

    @property
    def rotation_matrix(self):
        r"""
        The rotation matrix.

        :type: ``(n_dims, n_dims)`` `ndarray`
        """
        return self.linear_component

    def set_rotation_matrix(self, value, skip_checks=False):
        r"""
        Sets the rotation matrix.

        Parameters
        ----------
        value : ``(n_dims, n_dims)`` `ndarray`
            The new rotation matrix.
        skip_checks : `bool`, optional
            If ``True`` avoid sanity checks on ``value`` for performance.
        """
        if not skip_checks:
            shape = value.shape
            if len(shape) != 2 and shape[0] != shape[1]:
                raise ValueError("You need to provide a square rotation matrix")
            # The update better be the same size
            elif self.n_dims != shape[0]:
                raise ValueError("Trying to update the rotation "
                                 "matrix to a different dimension")
            # TODO actually check I am a valid rotation
            # TODO slightly dodgy here accessing _h_matrix
        self._h_matrix[:-1, :-1] = value

    def _transform_str(self):
        axis, radians_of_rotation = self.axis_and_angle_of_rotation()
        if axis is None:
            return "NO OP"
        degrees_of_rotation = np.rad2deg(radians_of_rotation)
        message = ('CCW Rotation of {:.1f} degrees '
                   'about {}'.format(degrees_of_rotation, axis))
        return message

    def axis_and_angle_of_rotation(self):
        r"""
        Abstract method for computing the axis and angle of rotation.

        Returns
        -------
        axis : ``(n_dims,)`` `ndarray`
            The unit vector representing the axis of rotation
        angle_of_rotation : `float`
            The angle in radians of the rotation about the axis. The angle is
            signed in a right handed sense.
        """
        if self.n_dims == 2:
            return self._axis_and_angle_of_rotation_2d()
        elif self.n_dims == 3:
            return self._axis_and_angle_of_rotation_3d()

    def _axis_and_angle_of_rotation_2d(self):
        r"""
        Decomposes this Rotation's rotation matrix into a angular rotation
        The rotation is considered in a right handed sense. The axis is, by
        definition, `[0, 0, 1]`.

        Returns
        -------
        axis : ``(2,)`` `ndarray`
            The vector representing the axis of rotation
        angle_of_rotation : `float`
            The angle in radians of the rotation about the axis. The angle is
            signed in a right handed sense.
        """
        axis = np.array([0, 0, 1])
        test_vector = np.array([1, 0])
        transformed_vector = np.dot(self.rotation_matrix,
                                    test_vector)
        angle_of_rotation = np.arccos(np.dot(transformed_vector, test_vector))
        return axis, angle_of_rotation

    def _axis_and_angle_of_rotation_3d(self):
        r"""
        Decomposes this 3D rotation's rotation matrix into a angular rotation
        about an axis. The rotation is considered in a right handed sense.

        Returns
        -------
        axis : ``(3,)`` `ndarray`
            A unit vector, the axis about which the rotation takes place
        angle_of_rotation : `float`
            The angle in radians of the rotation about the `axis`.
            The angle is signed in a right handed sense.

        References
        ----------
        .. [1] http://en.wikipedia.org/wiki/Rotation_matrix#Determining_the_axis
        """
        eval_, evec = np.linalg.eig(self.rotation_matrix)
        real_eval_mask = np.isreal(eval_)
        real_eval = np.real(eval_[real_eval_mask])
        evec_with_real_eval = np.real_if_close(evec[:, real_eval_mask])
        error = 1e-7
        below_margin = np.abs(real_eval) < (1 + error)
        above_margin = (1 - error) < np.abs(real_eval)
        re_unit_eval_mask = np.logical_and(below_margin, above_margin)
        evec_with_real_unitary_eval = evec_with_real_eval[:, re_unit_eval_mask]
        # all the eigenvectors with real unitary eigenvalues are now all
        # equally 'valid' if multiple remain that probably means that this
        # rotation is actually a no op (i.e. rotate by 360 degrees about any
        #  axis is an invariant transform) but need to check this. For now,
        # just take the first
        if evec_with_real_unitary_eval.shape[1] != 1:
            # TODO confirm that multiple eigenvalues of 1 means the rotation
            #  does nothing
            return None, None
        axis = evec_with_real_unitary_eval[:, 0]
        axis /= np.sqrt((axis ** 2).sum())  # normalize to unit vector
        # to find the angle of rotation, build a new unit vector perpendicular
        # to the axis, and see how it rotates
        axis_temp_vector = axis - np.random.rand(axis.size)
        perpendicular_vector = np.cross(axis, axis_temp_vector)
        perpendicular_vector /= np.sqrt((perpendicular_vector ** 2).sum())
        transformed_vector = np.dot(self.rotation_matrix,
                                    perpendicular_vector)
        angle_of_rotation = np.arccos(
            np.dot(transformed_vector, perpendicular_vector))
        chirality_of_rotation = np.dot(axis, np.cross(perpendicular_vector,
                                                      transformed_vector))
        if chirality_of_rotation < 0:
            angle_of_rotation *= -1.0
        return axis, angle_of_rotation

    @property
    def n_parameters(self):
        r"""
        Number of parameters of Rotation. Only 3D rotations are currently
        supported.

        Returns
        -------
        n_parameters : `int`
            The transform parameters. Only 3D rotations are currently
            supported which are parametrized with quaternions.

        Raises
        ------
        DimensionalityError, NotImplementedError
            Non-3D Rotations are not yet vectorizable
        """
        if self.n_dims == 3:
            # Quaternion parameters
            return 4
        else:
            raise NotImplementedError("Non-3D Rotations are not yet "
                                      "vectorizable")

    def _as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as quaternions. Only 3D transforms are currently
        supported.

        Returns
        -------
        q : ``(4,)`` `ndarray`
            The 4 quaternion parameters.

        Raises
        ------
        DimensionalityError, NotImplementedError
            Non-3D Rotations are not yet vectorizable
        """
        if self.n_dims == 3:
            m00 = self.h_matrix[0, 0]
            m01 = self.h_matrix[0, 1]
            m02 = self.h_matrix[0, 2]
            m10 = self.h_matrix[1, 0]
            m11 = self.h_matrix[1, 1]
            m12 = self.h_matrix[1, 2]
            m20 = self.h_matrix[2, 0]
            m21 = self.h_matrix[2, 1]
            m22 = self.h_matrix[2, 2]
            # symmetric matrix K
            K = np.array([[m00-m11-m22, 0.0,         0.0,         0.0],
                          [m01+m10,     m11-m00-m22, 0.0,         0.0],
                          [m02+m20,     m12+m21,     m22-m00-m11, 0.0],
                          [m21-m12,     m02-m20,     m10-m01,     m00+m11+m22]])
            K /= 3.0
            # Quaternion is eigenvector of K that corresponds to largest
            # eigenvalue
            w, V = np.linalg.eigh(K)
            q = V[[3, 0, 1, 2], np.argmax(w)]
            if q[0] < 0.0:
                q = -q
            return q
        else:
            raise NotImplementedError("Non-3D Rotations are not yet "
                                      "vectorizable")

    def _from_vector_inplace(self, p):
        r"""
        Returns an instance of the transform from the given parameters
        expressed in quaternions. Currently only 3D rotations are supported.

        Parameters
        ----------
        p : ``(4,)`` `ndarray`
            The array of quaternion parameters.

        Returns
        -------
        transform : :map:`Rotation`
            The transform initialised to the given parameters.

        Raises
        ------
        DimensionalityError, NotImplementedError
            Non-3D Rotations are not yet vectorizable
        ValueError
            Expected 4 quaternion parameters; got {} instead.
        """
        if self.n_dims == 3:
            if len(p) == 4:
                n = np.dot(p, p)
                # epsilon for testing whether a number is close to zero
                if n < np.finfo(float).eps * 4.0:
                    return np.identity(4)
                p = p * np.sqrt(2.0 / n)
                p = np.outer(p, p)
                rotation = np.array(
                    [[1.-p[2, 2]-p[3, 3],    p[1, 2]-p[3, 0],    p[1, 3]+p[2, 0]],
                     [   p[1, 2]+p[3, 0], 1.-p[1, 1]-p[3, 3],    p[2, 3]-p[1, 0]],
                     [   p[1, 3]-p[2, 0],    p[2, 3]+p[1, 0], 1.-p[1, 1]-p[2, 2]]])
                self.set_rotation_matrix(rotation, skip_checks=True)
            else:
                raise ValueError("Expected 4 quaternion parameters; got {} "
                                 "instead.".format(len(p)))
        else:
            raise NotImplementedError("Non-3D rotations are not yet "
                                      "vectorizable")

    @property
    def composes_inplace_with(self):
        r"""
        :class:`Rotation` can swallow composition with any other
        :class:`Rotation`.
        """
        return Rotation

    def pseudoinverse(self):
        r"""
        The inverse rotation matrix.

        :type: :class:`Rotation`
        """
        return Rotation(np.linalg.inv(self.rotation_matrix), skip_checks=True)


class AlignmentRotation(HomogFamilyAlignment, Rotation):
    r"""
    Constructs an :class:`Rotation` by finding the optimal rotation transform to
    align `source` to `target`.

    Parameters
    ----------
    source : :map:`PointCloud`
        The source pointcloud instance used in the alignment
    target : :map:`PointCloud`
        The target pointcloud instance used in the alignment
    allow_mirror : `bool`, optional
        If ``True``, the Kabsch algorithm check is not performed, and mirroring
        of the Rotation matrix is permitted.
    """

    def __init__(self, source, target, allow_mirror=False):
        HomogFamilyAlignment.__init__(self, source, target)
        Rotation.__init__(self, optimal_rotation_matrix(
            source, target, allow_mirror=allow_mirror))
        self.allow_mirror = allow_mirror

    def set_rotation_matrix(self, value, skip_checks=False):
        r"""
        Sets the rotation matrix.

        Parameters
        ----------
        value : ``(n_dims, n_dims)`` `ndarray`
            The new rotation matrix.
        skip_checks : `bool`, optional
            If ``True`` avoid sanity checks on ``value`` for performance.
        """
        Rotation.set_rotation_matrix(self, value, skip_checks=skip_checks)
        self._sync_target_from_state()

    def _sync_state_from_target(self):
        r = optimal_rotation_matrix(self.source, self.target,
                                    allow_mirror=self.allow_mirror)
        Rotation.set_rotation_matrix(self, r, skip_checks=True)

    def as_non_alignment(self):
        r"""
        Returns a copy of this rotation without its alignment nature.

        Returns
        -------
        transform : :map:`Rotation`
            A version of this rotation with the same transform behavior but
            without the alignment logic.
        """
        return Rotation(self.rotation_matrix, skip_checks=True)



from __future__ import division
from warnings import warn
import numpy as np

binary_erosion = None  # expensive, from scipy.ndimage
binary_dilation = None  # expensive, from scipy.ndimage

from menpo.base import MenpoDeprecationWarning, copy_landmarks_and_path
from menpo.transform import Translation
from menpo.visualize.base import ImageViewer

from .base import Image
from .boolean import BooleanImage


class OutOfMaskSampleError(ValueError):
    r"""
    Exception that is thrown when an attempt is made to sample an MaskedImage
    in an area that is masked out (where the mask is ``False``).

    Parameters
    ----------
    sampled_mask : `bool ndarray`
        The sampled mask, ``True`` where the image's mask was ``True`` and
        ``False`` otherwise. Useful for masking out the sampling array.
    sampled_values : `ndarray`
        The sampled values, no attempt at masking is made.
    """
    def __init__(self, sampled_mask, sampled_values):
        super(OutOfMaskSampleError, self).__init__()
        self.sampled_mask = sampled_mask
        self.sampled_values = sampled_values


class MaskedImage(Image):
    r"""
    Represents an `n`-dimensional `k`-channel image, which has a mask.
    Images can be masked in order to identify a region of interest. All
    images implicitly have a mask that is defined as the the entire image.
    The mask is an instance of :map:`BooleanImage`.

    Parameters
    ----------
    image_data :  ``(C, M, N ..., Q)`` `ndarray`
        The pixel data for the image, where the first axis represents the
        number of channels.
    mask : ``(M, N)`` `bool ndarray` or :map:`BooleanImage`, optional
        A binary array representing the mask. Must be the same
        shape as the image. Only one mask is supported for an image (so the
        mask is applied to every channel equally).
    copy: `bool`, optional
        If ``False``, the ``image_data`` will not be copied on assignment. If a
        mask is provided, this also won't be copied. In general this should only
        be used if you know what you are doing.

    Raises
    ------
    ValueError
        Mask is not the same shape as the image
    """
    def __init__(self, image_data, mask=None, copy=True):
        super(MaskedImage, self).__init__(image_data, copy=copy)
        if mask is not None:
            # Check if we need to create a BooleanImage or not
            if not isinstance(mask, BooleanImage):
                # So it's a numpy array.
                mask_image = BooleanImage(mask, copy=copy)
            else:
                # It's a BooleanImage object.
                if copy:
                    mask = mask.copy()
                mask_image = mask
            if mask_image.shape == self.shape:
                self.mask = mask_image
            else:
                raise ValueError("Trying to construct a Masked Image of "
                                 "shape {} with a Mask of differing "
                                 "shape {}".format(self.shape,
                                                   mask.shape))
        else:
            # no mask provided - make the default.
            self.mask = BooleanImage.init_blank(self.shape, fill=True)

    @classmethod
    def init_blank(cls, shape, n_channels=1, fill=0, dtype=np.float, mask=None):
        r"""Generate a blank masked image

        Parameters
        ----------
        shape : `tuple` or `list`
            The shape of the image. Any floating point values are rounded up
            to the nearest integer.
        n_channels: `int`, optional
            The number of channels to create the image with.
        fill : `int`, optional
            The value to fill all pixels with.
        dtype: `numpy datatype`, optional
            The datatype of the image.
        mask: ``(M, N)`` `bool ndarray` or :map:`BooleanImage`
            An optional mask that can be applied to the image. Has to have a
            shape equal to that of the image.

        Notes
        -----
        Subclasses of :map:`MaskedImage` need to overwrite this method and
        explicitly call this superclass method

        ::

            super(SubClass, cls).init_blank(shape,**kwargs)

        in order to appropriately propagate the subclass type to ``cls``.

        Returns
        -------
        blank_image : :map:`MaskedImage`
            A new masked image of the requested size.
        """
        # Ensure that the '+' operator means concatenate tuples
        shape = tuple(np.ceil(shape).astype(np.int))
        if fill == 0:
            pixels = np.zeros((n_channels,) + shape, dtype=dtype)
        else:
            pixels = np.ones((n_channels,) + shape, dtype=dtype) * fill
        return cls(pixels, copy=False, mask=mask)

    @classmethod
    def init_from_channels_at_back(cls, pixels, mask=None):
        r"""
        Create an Image from a set of pixels where the channels axis is on
        the last axis (the back). This is common in other frameworks, and
        therefore this method provides a convenient means of creating a menpo
        Image from such data. Note that a copy is always created due to the
        need to rearrange the data.

        Parameters
        ----------
        pixels : ``(M, N ..., Q, C)`` `ndarray`
            Array representing the image pixels, with the last axis being
            channels.
        mask : ``(M, N)`` `bool ndarray` or :map:`BooleanImage`, optional
            A binary array representing the mask. Must be the same
            shape as the image. Only one mask is supported for an image (so the
            mask is applied to every channel equally).

        Returns
        -------
        image : :map:`Image`
            A new image from the given pixels, with the FIRST axis as the
            channels.
        """
        im = Image.init_from_channels_at_back(pixels)
        if mask is not None:
            mask = mask.copy()
        return MaskedImage(im.pixels, mask=mask, copy=False)

    @classmethod
    def init_from_pointcloud(cls, pointcloud, group=None, boundary=0,
                             constrain_mask=True, n_channels=1, fill=0,
                             dtype=np.float):
        r"""
        Create an Image that is big enough to contain the given pointcloud.
        The pointcloud will be translated to the origin and then translated
        according to its bounds in order to fit inside the new image.
        An optional boundary can be provided in order to increase the space
        around the boundary of the pointcloud. The boundary will be added
        to *all sides of the image* and so a boundary of 5 provides 10 pixels
        of boundary total for each dimension.

        By default, the mask will be constrained to the convex hull of the
        provided pointcloud.

        Parameters
        ----------
        pointcloud : :map:`PointCloud`
            Pointcloud to place inside the newly created image.
        group : `str`, optional
            If ``None``, the pointcloud will only be used to create the image.
            If a `str` then the pointcloud will be attached as a landmark
            group to the image, with the given string as key.
        boundary : `float`
            A optional padding distance that is added to the pointcloud bounds.
            Default is ``0``, meaning the max/min of tightest possible
            containing image is returned.
        n_channels : `int`, optional
            The number of channels to create the image with.
        fill : `int`, optional
            The value to fill all pixels with.
        dtype : numpy data type, optional
            The data type of the image.
        constrain_mask : `bool`, optional
            If ``True``, the mask will be constrained to the convex hull
            of the provided pointcloud. If ``False``, the mask will be all
            ``True``.

        Returns
        -------
        image : :map:`MaskedImage`
            A new image with the same size as the given pointcloud, optionally
            with the pointcloud attached as landmarks and the mask constrained
            to the convex hull of the pointcloud.
        """
        # Translate pointcloud to the origin
        minimum = pointcloud.bounds(boundary=boundary)[0]
        origin_pc = Translation(-minimum).apply(pointcloud)
        image_shape = origin_pc.range(boundary=boundary)
        if constrain_mask:
            new_mask = BooleanImage.init_from_pointcloud(
                origin_pc, group=None, boundary=boundary, constrain=True,
                fill=False)
        else:
            new_mask = None

        new_image = cls.init_blank(image_shape, n_channels=n_channels,
                                   fill=fill, dtype=dtype, mask=new_mask)
        if group is not None:
            new_image.landmarks[group] = origin_pc
        return new_image

    def as_unmasked(self, copy=True, fill=None):
        r"""
        Return a copy of this image without the masking behavior.

        By default the mask is simply discarded. However, there is an optional
        kwarg, ``fill``, that can be set which will fill the **non-masked**
        areas with the given value.

        Parameters
        ----------
        copy : `bool`, optional
            If ``False``, the produced :map:`Image` will share pixels with
            ``self``. Only suggested to be used for performance.
        fill : `float` or ``(n_channels,)`` iterable or ``None``, optional
            If ``None`` the mask is simply discarded. If a scalar or iterable,
            the *unmasked* regions are filled with the given value.

        Returns
        -------
        image : :map:`Image`
            An image with the same pixels and landmarks as this one, but with
            no mask.
        """
        img = Image(self.pixels, copy=copy)
        if fill is not None:
            if not np.isscalar(fill):
                fill = np.array(fill).reshape(self.n_channels, -1)
            img.pixels[..., ~self.mask.mask] = fill
        return copy_landmarks_and_path(self, img)

    def n_true_pixels(self):
        r"""
        The number of ``True`` values in the mask.

        :type: `int`
        """
        return self.mask.n_true()

    def n_false_pixels(self):
        r"""
        The number of ``False`` values in the mask.

        :type: `int`
        """
        return self.mask.n_false()

    def n_true_elements(self):
        r"""
        The number of ``True`` elements of the image over all the channels.

        :type: `int`
        """
        return self.n_true_pixels() * self.n_channels

    def n_false_elements(self):
        r"""
        The number of ``False`` elements of the image over all the channels.

        :type: `int`
        """
        return self.n_false_pixels() * self.n_channels

    def indices(self):
        r"""
        Return the indices of all true pixels in this image.

        :type: ``(n_dims, n_true_pixels)`` `ndarray`
        """
        return self.mask.true_indices()

    def masked_pixels(self):
        r"""
        Get the pixels covered by the `True` values in the mask.

        :type: ``(n_channels, mask.n_true)`` `ndarray`
        """
        if self.mask.all_true():
            return self.pixels
        return self.pixels[..., self.mask.mask]

    def set_masked_pixels(self, pixels, copy=True):
        r"""
        Deprecated - please use the equivalent ``from_vector``
        """
        warn('This method is no longer supported and will be removed in a '
             'future version of Menpo. '
             'Use .from_vector() instead.',
             MenpoDeprecationWarning)
        self._set_masked_pixels(pixels, copy=copy)

    # TODO: Replace _from_vector_inplace with this.
    def _set_masked_pixels(self, pixels, copy=True):
        r"""
        Update the masked pixels only to new values.

        Parameters
        ----------
        pixels: `ndarray`
            The new pixels to set.
        copy: `bool`, optional
            If ``False`` a copy will be avoided in assignment. This can only
            happen if the mask is all ``True`` - in all other cases it will
            raise a warning.

        Raises
        ------
        Warning
            If the ``copy=False`` flag cannot be honored.
        """
        if self.mask.all_true():
            # reshape the vector into the image again
            pixels = pixels.reshape((self.n_channels,) + self.shape)
            if not copy:
                if not pixels.flags.c_contiguous:
                    warn('The copy flag was NOT honoured. A copy HAS been '
                         'made. Copy can only be avoided if MaskedImage has '
                         'an all_true mask and the pixels provided are '
                         'C-contiguous.')
                    pixels = pixels.copy()
            else:
                pixels = pixels.copy()
            self.pixels = pixels
        else:
            self.pixels[..., self.mask.mask] = pixels
            # oh dear, couldn't avoid a copy. Did the user try to?
            if not copy:
                warn('The copy flag was NOT honoured. A copy HAS been made. '
                     'copy can only be avoided if MaskedImage has an all_true '
                     'mask.')

    def __str__(self):
        return ('{} {}D MaskedImage with {} channels. '
                'Attached mask {:.1%} true'.format(
            self._str_shape(), self.n_dims, self.n_channels,
            self.mask.proportion_true()))

    def _as_vector(self, keep_channels=False):
        r"""
        Convert image to a vectorized form. Note that the only pixels
        returned here are from the masked region on the image.

        Parameters
        ----------
        keep_channels : `bool`, optional

            ========== =================================
            Value      Return shape
            ========== =================================
            ``True``     ``(mask.n_true, n_channels)``
            ``False``    ``(mask.n_true * n_channels,)``
            ========== =================================

        Returns
        -------
        vectorized_image : (shape given by ``keep_channels``) `ndarray`
            Vectorized image
        """
        if keep_channels:
            return self.masked_pixels().reshape([self.n_channels, -1])
        else:
            return self.masked_pixels().ravel()

    def from_vector(self, vector, n_channels=None):
        r"""
        Takes a flattened vector and returns a new image formed by reshaping
        the vector to the correct pixels and channels. Note that the only
        region of the image that will be filled is the masked region.

        On masked images, the vector is always copied.

        The ``n_channels`` argument is useful for when we want to add an extra
        channel to an image but maintain the shape. For example, when
        calculating the gradient.

        Note that landmarks are transferred in the process.

        Parameters
        ----------
        vector : ``(n_pixels,)``
            A flattened vector of all pixels and channels of an image.
        n_channels : `int`, optional
            If given, will assume that vector is the same shape as this image,
            but with a possibly different number of channels.

        Returns
        -------
        image : :class:`MaskedImage`
            New image of same shape as this image and the number of
            specified channels.
        """
        # This is useful for when we want to add an extra channel to an image
        # but maintain the shape. For example, when calculating the gradient
        n_channels = self.n_channels if n_channels is None else n_channels
        # Creates zeros of size (n_channels x M x N x ...)
        if self.mask.all_true():
            # we can just reshape the array!
            image_data = vector.reshape(((n_channels,) + self.shape))
        else:
            image_data = np.zeros((n_channels,) + self.shape,
                                  dtype=vector.dtype)
            pixels_per_channel = vector.reshape((n_channels, -1))
            image_data[..., self.mask.mask] = pixels_per_channel
        new_image = MaskedImage(image_data, mask=self.mask)
        return copy_landmarks_and_path(self, new_image)

    def _from_vector_inplace(self, vector, copy=True):
        r"""
        Takes a flattened vector and updates this image by reshaping
        the vector to the correct pixels and channels. Note that the only
        region of the image that will be filled is the masked region.

        Parameters
        ----------
        vector : ``(n_parameters,)``
            A flattened vector of all pixels and channels of an image.
        copy : `bool`, optional
            If ``False``, the vector will be set as the pixels with no copy
            made.
            If ``True`` a copy of the vector is taken.

        Raises
        ------
        Warning
            If ``copy=False`` cannot be honored.
        """
        self._set_masked_pixels(vector.reshape((self.n_channels, -1)),
                                copy=copy)

    def _view_2d(self, figure_id=None, new_figure=False, channels=None,
                 masked=True, interpolation='bilinear', cmap_name=None,
                 alpha=1., render_axes=False, axes_font_name='sans-serif',
                 axes_font_size=10, axes_font_style='normal',
                 axes_font_weight='normal', axes_x_limits=None,
                 axes_y_limits=None, axes_x_ticks=None, axes_y_ticks=None,
                 figure_size=(7, 7)):
        r"""
        View the image using the default image viewer. This method will appear
        on the Image as ``view`` if the Image is 2D.

        Returns
        -------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        channels : `int` or `list` of `int` or ``all`` or ``None``
            If `int` or `list` of `int`, the specified channel(s) will be
            rendered. If ``all``, all the channels will be rendered in subplots.
            If ``None`` and the image is RGB, it will be rendered in RGB mode.
            If ``None`` and the image is not RGB, it is equivalent to ``all``.
        masked : `bool`, optional
            If ``True``, only the masked pixels will be rendered.
        interpolation : See Below, optional
            The interpolation used to render the image. For example, if
            ``bilinear``, the image will be smooth and if ``nearest``, the
            image will be pixelated.
            Example options ::

                {none, nearest, bilinear, bicubic, spline16, spline36,
                hanning, hamming, hermite, kaiser, quadric, catrom, gaussian,
                bessel, mitchell, sinc, lanczos}

        cmap_name: `str`, optional,
            If ``None``, single channel and three channel images default
            to greyscale and rgb colormaps respectively.
        alpha : `float`, optional
            The alpha blending value, between 0 (transparent) and 1 (opaque).
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the x axis. If `float`, then it sets padding on the
            right and left of the Image as a percentage of the Image's width. If
            `tuple` or `list`, then it defines the axis limits. If ``None``, then
            the limits are set automatically.
        axes_y_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the y axis. If `float`, then it sets padding on the
            top and bottom of the Image as a percentage of the Image's height. If
            `tuple` or `list`, then it defines the axis limits. If ``None``, then
            the limits are set automatically.
        axes_x_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the x axis.
        axes_y_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the y axis.
        figure_size : (`float`, `float`) `tuple` or ``None``, optional
            The size of the figure in inches.

        Raises
        ------
        ValueError
            If Image is not 2D
        """
        mask = self.mask.mask if masked else None
        return ImageViewer(figure_id, new_figure, self.n_dims,
                           self.pixels, channels=channels,
                           mask=mask).render(
            interpolation=interpolation, cmap_name=cmap_name, alpha=alpha,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
            axes_y_limits=axes_y_limits, axes_x_ticks=axes_x_ticks,
            axes_y_ticks=axes_y_ticks, figure_size=figure_size)

    def _view_landmarks_2d(self, channels=None, masked=True, group=None,
                           with_labels=None, without_labels=None,
                           figure_id=None, new_figure=False,
                           interpolation='bilinear', cmap_name=None, alpha=1.,
                           render_lines=True, line_colour=None, line_style='-',
                           line_width=1, render_markers=True, marker_style='o',
                           marker_size=5, marker_face_colour=None,
                           marker_edge_colour=None, marker_edge_width=1.,
                           render_numbering=False,
                           numbers_horizontal_align='center',
                           numbers_vertical_align='bottom',
                           numbers_font_name='sans-serif', numbers_font_size=10,
                           numbers_font_style='normal',
                           numbers_font_weight='normal',
                           numbers_font_colour='k', render_legend=False,
                           legend_title='', legend_font_name='sans-serif',
                           legend_font_style='normal', legend_font_size=10,
                           legend_font_weight='normal',
                           legend_marker_scale=None,
                           legend_location=2, legend_bbox_to_anchor=(1.05, 1.),
                           legend_border_axes_pad=None, legend_n_columns=1,
                           legend_horizontal_spacing=None,
                           legend_vertical_spacing=None, legend_border=True,
                           legend_border_padding=None, legend_shadow=False,
                           legend_rounded_corners=False, render_axes=False,
                           axes_font_name='sans-serif', axes_font_size=10,
                           axes_font_style='normal', axes_font_weight='normal',
                           axes_x_limits=None, axes_y_limits=None,
                           axes_x_ticks=None, axes_y_ticks=None,
                           figure_size=(7, 7)):
        """
        Visualize the landmarks. This method will appear on the Image as
        ``view_landmarks`` if the Image is 2D.

        Parameters
        ----------
        channels : `int` or `list` of `int` or ``all`` or ``None``
            If `int` or `list` of `int`, the specified channel(s) will be
            rendered. If ``all``, all the channels will be rendered in subplots.
            If ``None`` and the image is RGB, it will be rendered in RGB mode.
            If ``None`` and the image is not RGB, it is equivalent to ``all``.
        masked : `bool`, optional
            If ``True``, only the masked pixels will be rendered.
        group : `str` or``None`` optionals
            The landmark group to be visualized. If ``None`` and there are more
            than one landmark groups, an error is raised.
        with_labels : ``None`` or `str` or `list` of `str`, optional
            If not ``None``, only show the given label(s). Should **not** be
            used with the ``without_labels`` kwarg.
        without_labels : ``None`` or `str` or `list` of `str`, optional
            If not ``None``, show all except the given label(s). Should **not**
            be used with the ``with_labels`` kwarg.
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        interpolation : See Below, optional
            The interpolation used to render the image. For example, if
            ``bilinear``, the image will be smooth and if ``nearest``, the
            image will be pixelated. Example options ::

                {none, nearest, bilinear, bicubic, spline16, spline36, hanning,
                hamming, hermite, kaiser, quadric, catrom, gaussian, bessel,
                mitchell, sinc, lanczos}

        cmap_name: `str`, optional,
            If ``None``, single channel and three channel images default
            to greyscale and rgb colormaps respectively.
        alpha : `float`, optional
            The alpha blending value, between 0 (transparent) and 1 (opaque).
        render_lines : `bool`, optional
            If ``True``, the edges will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        line_style : ``{-, --, -., :}``, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers. Example options ::

                {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_numbering : `bool`, optional
            If ``True``, the landmarks will be numbered.
        numbers_horizontal_align : ``{center, right, left}``, optional
            The horizontal alignment of the numbers' texts.
        numbers_vertical_align : ``{center, top, bottom, baseline}``, optional
            The vertical alignment of the numbers' texts.
        numbers_font_name : See Below, optional
            The font of the numbers. Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        numbers_font_size : `int`, optional
            The font size of the numbers.
        numbers_font_style : ``{normal, italic, oblique}``, optional
            The font style of the numbers.
        numbers_font_weight : See Below, optional
            The font weight of the numbers.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        numbers_font_colour : See Below, optional
            The font colour of the numbers.
            Example options ::

                {r, g, b, c, m, k, w}
                or
                (3, ) ndarray

        render_legend : `bool`, optional
            If ``True``, the legend will be rendered.
        legend_title : `str`, optional
            The title of the legend.
        legend_font_name : See below, optional
            The font of the legend. Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        legend_font_style : ``{normal, italic, oblique}``, optional
            The font style of the legend.
        legend_font_size : `int`, optional
            The font size of the legend.
        legend_font_weight : See Below, optional
            The font weight of the legend.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold, demibold, demi, bold, heavy, extra bold, black}

        legend_marker_scale : `float`, optional
            The relative size of the legend markers with respect to the original
        legend_location : `int`, optional
            The location of the legend. The predefined values are:

            =============== ==
            'best'          0
            'upper right'   1
            'upper left'    2
            'lower left'    3
            'lower right'   4
            'right'         5
            'center left'   6
            'center right'  7
            'lower center'  8
            'upper center'  9
            'center'        10
            =============== ==

        legend_bbox_to_anchor : (`float`, `float`) `tuple`, optional
            The bbox that the legend will be anchored.
        legend_border_axes_pad : `float`, optional
            The pad between the axes and legend border.
        legend_n_columns : `int`, optional
            The number of the legend's columns.
        legend_horizontal_spacing : `float`, optional
            The spacing between the columns.
        legend_vertical_spacing : `float`, optional
            The vertical space between the legend entries.
        legend_border : `bool`, optional
            If ``True``, a frame will be drawn around the legend.
        legend_border_padding : `float`, optional
            The fractional whitespace inside the legend border.
        legend_shadow : `bool`, optional
            If ``True``, a shadow will be drawn behind legend.
        legend_rounded_corners : `bool`, optional
            If ``True``, the frame's corners will be rounded (fancybox).
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes. Example options ::

                {serif, sans-serif, cursive, fantasy, monospace}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : ``{normal, italic, oblique}``, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {ultralight, light, normal, regular, book, medium, roman,
                semibold,demibold, demi, bold, heavy, extra bold, black}

        axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
            The limits of the x axis. If `float`, then it sets padding on the
            right and left of the Image as a percentage of the Image's width. If
            `tuple` or `list`, then it defines the axis limits. If ``None``, then
            the limits are set automatically.
        axes_y_limits : (`float`, `float`) `tuple` or ``None``, optional
            The limits of the y axis. If `float`, then it sets padding on the
            top and bottom of the Image as a percentage of the Image's height. If
            `tuple` or `list`, then it defines the axis limits. If ``None``, then
            the limits are set automatically.
        axes_x_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the x axis.
        axes_y_ticks : `list` or `tuple` or ``None``, optional
            The ticks of the y axis.
        figure_size : (`float`, `float`) `tuple` or ``None`` optional
            The size of the figure in inches.

        Raises
        ------
        ValueError
            If both ``with_labels`` and ``without_labels`` are passed.
        ValueError
            If the landmark manager doesn't contain the provided group label.
        """
        from menpo.visualize import view_image_landmarks
        return view_image_landmarks(
            self, channels, masked, group, with_labels, without_labels,
            figure_id, new_figure, interpolation, cmap_name, alpha,
            render_lines, line_colour, line_style, line_width, render_markers,
            marker_style, marker_size, marker_face_colour, marker_edge_colour,
            marker_edge_width, render_numbering, numbers_horizontal_align,
            numbers_vertical_align, numbers_font_name, numbers_font_size,
            numbers_font_style, numbers_font_weight, numbers_font_colour,
            render_legend, legend_title, legend_font_name, legend_font_style,
            legend_font_size, legend_font_weight, legend_marker_scale,
            legend_location, legend_bbox_to_anchor, legend_border_axes_pad,
            legend_n_columns, legend_horizontal_spacing,
            legend_vertical_spacing, legend_border, legend_border_padding,
            legend_shadow, legend_rounded_corners, render_axes, axes_font_name,
            axes_font_size, axes_font_style, axes_font_weight, axes_x_limits,
            axes_y_limits, axes_x_ticks, axes_y_ticks, figure_size)

    def crop_to_true_mask(self, boundary=0, constrain_to_boundary=True,
                          return_transform=False):
        r"""
        Crop this image to be bounded just the `True` values of it's mask.

        Parameters
        ----------
        boundary : `int`, optional
            An extra padding to be added all around the true mask region.
        constrain_to_boundary : `bool`, optional
            If ``True`` the crop will be snapped to not go beyond this images
            boundary. If ``False``, an :map:`ImageBoundaryError` will be raised
            if an attempt is made to go beyond the edge of the image. Note that
            is only possible if ``boundary != 0``.
        return_transform : `bool`, optional
            If ``True``, then the :map:`Transform` object that was used to
            perform the cropping is also returned.

        Returns
        -------
        cropped_image : ``type(self)``
            A copy of this image, cropped to the true mask.
        transform : :map:`Transform`
            The transform that was used. It only applies if
            `return_transform` is ``True``.

        Raises
        ------
        ImageBoundaryError
            Raised if 11constrain_to_boundary=False`1, and an attempt is
            made to crop the image in a way that violates the image bounds.
        """
        min_indices, max_indices = self.mask.bounds_true(
            boundary=boundary, constrain_to_bounds=False)
        # no point doing the bounds check twice - let the crop do it only.
        return self.crop(min_indices, max_indices,
                         constrain_to_boundary=constrain_to_boundary,
                         return_transform=return_transform)

    def sample(self, points_to_sample, order=1, mode='constant', cval=0.0,
               verify_mask=False):
        r"""
        Sample this image at the given sub-pixel accurate points. The input
        PointCloud should have the same number of dimensions as the image e.g.
        a 2D PointCloud for a 2D multi-channel image. A numpy array will be
        returned the has the values for every given point across each channel
        of the image.

        If verify_mask is True and the points to sample are *outside* of the
        mask (fall on a ``False`` value in the mask), an exception is raised.
        This exception contains the information of which points were outside
        of the mask (``False``) and *also* returns the sampled points. Note this
        is more expensive and thus may be disabled by setting verify_mask
        to False.

        Parameters
        ----------
        points_to_sample : :map:`PointCloud`
            Array of points to sample from the image. Should be
            `(n_points, n_dims)`
        order : `int`, optional
            The order of interpolation. The order has to be in the range [0,5].
            See warp_to_shape for more information.
        mode : ``{constant, nearest, reflect, wrap}``, optional
            Points outside the boundaries of the input are filled according
            to the given mode.
        cval : `float`, optional
            Used in conjunction with mode ``constant``, the value outside
            the image boundaries.
        verify_mask : `bool`, optional
            If True, also sample the mask at the given points and check the
            mask is valid at all points. Note that sampling masks with higher
            order splines may cause interpolated mask values that are rounded
            to zero and thus cause false positives.

        Returns
        -------
        sampled_pixels : (`n_points`, `n_channels`) `ndarray`
            The interpolated values taken across every channel of the image.

        Raises
        ------
        OutOfMaskSampleError
            One of the points to sample was outside of the valid area of the
            mask (``False`` in the mask). This exception contains both the
            mask of valid sample points, **as well as** the sampled points
            themselves, in case you want to ignore the error. Only raised
            if verify_mask is True.
        """
        sampled_values = Image.sample(self, points_to_sample, order=order,
                                      mode=mode, cval=cval)
        if verify_mask:
            sampled_mask = self.mask.sample(points_to_sample, mode=mode,
                                            cval=cval)
            if not np.all(sampled_mask):
                raise OutOfMaskSampleError(sampled_mask, sampled_values)
        return sampled_values

    # noinspection PyMethodOverriding
    def warp_to_mask(self, template_mask, transform, warp_landmarks=False,
                     order=1, mode='constant', cval=0., batch_size=None,
                     return_transform=False):
        r"""
        Warps this image into a different reference space.

        Parameters
        ----------
        template_mask : :map:`BooleanImage`
            Defines the shape of the result, and what pixels should be sampled.
        transform : :map:`Transform`
            Transform **from the template space back to this image**.
            Defines, for each pixel location on the template, which pixel
            location should be sampled from on this image.
        warp_landmarks : `bool`, optional
            If ``True``, result will have the same landmark dictionary
            as ``self``, but with each landmark updated to the warped position.
        order : `int`, optional
            The order of interpolation. The order has to be in the range [0,5]

            ========= =====================
            Order     Interpolation
            ========= =====================
            0         Nearest-neighbor
            1         Bi-linear *(default)*
            2         Bi-quadratic
            3         Bi-cubic
            4         Bi-quartic
            5         Bi-quintic
            ========= =====================

        mode : ``{constant, nearest, reflect, wrap}``, optional
            Points outside the boundaries of the input are filled according
            to the given mode.
        cval : `float`, optional
            Used in conjunction with mode ``constant``, the value outside
            the image boundaries.
        batch_size : `int` or ``None``, optional
            This should only be considered for large images. Setting this
            value can cause warping to become much slower, particular for
            cached warps such as Piecewise Affine. This size indicates
            how many points in the image should be warped at a time, which
            keeps memory usage low. If ``None``, no batching is used and all
            points are warped at once.
        return_transform : `bool`, optional
            This argument is for internal use only. If ``True``, then the
            :map:`Transform` object is also returned.

        Returns
        -------
        warped_image : ``type(self)``
            A copy of this image, warped.
        transform : :map:`Transform`
            The transform that was used. It only applies if
            `return_transform` is ``True``.
        """
        # call the super variant and get ourselves a MaskedImage back
        # with a blank mask
        warped_image = Image.warp_to_mask(self, template_mask, transform,
                                          warp_landmarks=warp_landmarks,
                                          order=order, mode=mode, cval=cval,
                                          batch_size=batch_size)
        # Set the template mask as our mask
        warped_image.mask = template_mask
        # optionally return the transform
        if return_transform:
            return warped_image, transform
        else:
            return warped_image

    # noinspection PyMethodOverriding
    def warp_to_shape(self, template_shape, transform, warp_landmarks=False,
                      order=1, mode='constant', cval=0., batch_size=None,
                      return_transform=False):
        """
        Return a copy of this :map:`MaskedImage` warped into a different
        reference space.

        Parameters
        ----------
        template_shape : `tuple` or `ndarray`
            Defines the shape of the result, and what pixel indices should be
            sampled (all of them).
        transform : :map:`Transform`
            Transform **from the template_shape space back to this image**.
            Defines, for each index on template_shape, which pixel location
            should be sampled from on this image.
        warp_landmarks : `bool`, optional
            If ``True``, result will have the same landmark dictionary
            as self, but with each landmark updated to the warped position.
        order : `int`, optional
            The order of interpolation. The order has to be in the range [0,5]

            ========= =====================
            Order     Interpolation
            ========= =====================
            0         Nearest-neighbor
            1         Bi-linear *(default)*
            2         Bi-quadratic
            3         Bi-cubic
            4         Bi-quartic
            5         Bi-quintic
            ========= =====================

        mode : ``{constant, nearest, reflect, wrap}``, optional
            Points outside the boundaries of the input are filled according
            to the given mode.
        cval : `float`, optional
            Used in conjunction with mode ``constant``, the value outside
            the image boundaries.
        batch_size : `int` or ``None``, optional
            This should only be considered for large images. Setting this
            value can cause warping to become much slower, particular for
            cached warps such as Piecewise Affine. This size indicates
            how many points in the image should be warped at a time, which
            keeps memory usage low. If ``None``, no batching is used and all
            points are warped at once.
        return_transform : `bool`, optional
            This argument is for internal use only. If ``True``, then the
            :map:`Transform` object is also returned.

        Returns
        -------
        warped_image : :map:`MaskedImage`
            A copy of this image, warped.
        transform : :map:`Transform`
            The transform that was used. It only applies if
            `return_transform` is ``True``.
        """
        # call the super variant and get ourselves an Image back
        warped_image = Image.warp_to_shape(self, template_shape, transform,
                                           warp_landmarks=warp_landmarks,
                                           order=order, mode=mode, cval=cval,
                                           batch_size=batch_size)
        # Warp the mask separately and reattach
        mask = self.mask.warp_to_shape(template_shape, transform,
                                       warp_landmarks=warp_landmarks,
                                       mode=mode, cval=cval)
        # efficiently turn the Image into a MaskedImage, attaching the
        # landmarks
        masked_warped_image = warped_image.as_masked(mask=mask, copy=False)
        if hasattr(warped_image, 'path'):
            masked_warped_image.path = warped_image.path
        # optionally return the transform
        if return_transform:
            return masked_warped_image, transform
        else:
            return masked_warped_image

    def normalize_std(self, mode='all', limit_to_mask=True):
        r"""
        Returns a copy of this image normalized such that it's pixel values
        have zero mean and unit variance.

        Parameters
        ----------
        mode : ``{all, per_channel}``, optional
            If ``all``, the normalization is over all channels. If
            ``per_channel``, each channel individually is mean centred and
            normalized in variance.
        limit_to_mask : `bool`, optional
            If ``True``, the normalization is only performed wrt the masked
            pixels.
            If ``False``, the normalization is wrt all pixels, regardless of
            their masking value.

        Returns
        -------
        image : ``type(self)``
            A copy of this image, normalized.
        """
        warn('This method is no longer supported and will be removed in a '
             'future version of Menpo. '
             'Use .normalize_std() instead (features package).',
             MenpoDeprecationWarning)

        return self._normalize(np.std, mode=mode,
                               limit_to_mask=limit_to_mask)

    def normalize_norm(self, mode='all', limit_to_mask=True, **kwargs):
        r"""
        Returns a copy of this image normalized such that it's pixel values
        have zero mean and its norm equals 1.

        Parameters
        ----------
        mode : ``{all, per_channel}``, optional
            If ``all``, the normalization is over all channels. If
            ``per_channel``, each channel individually is mean centred and
            normalized in variance.
        limit_to_mask : `bool`, optional
            If ``True``, the normalization is only performed wrt the masked
            pixels.
            If ``False``, the normalization is wrt all pixels, regardless of
            their masking value.

        Returns
        -------
        image : ``type(self)``
            A copy of this image, normalized.
        """
        warn('This method is no longer supported and will be removed in a '
             'future version of Menpo. '
             'Use .normalize_norm() instead (features package).',
             MenpoDeprecationWarning)

        def scale_func(pixels, axis=None):
            return np.linalg.norm(pixels, axis=axis, **kwargs)

        return self._normalize(scale_func, mode=mode,
                               limit_to_mask=limit_to_mask)

    def _normalize(self, scale_func, mode='all', limit_to_mask=True):
        from menpo.feature import normalize
        if limit_to_mask:
            pixels = self
        else:
            pixels = self.as_unmasked(copy=False)

        new_img = normalize(pixels, scale_func=scale_func, mode=mode)

        if limit_to_mask:
            return new_img
        else:
            return new_img.as_masked(copy=False, mask=self.mask.copy())

    def constrain_mask_to_landmarks(self, group=None, batch_size=None,
                                    point_in_pointcloud='pwa'):
        r"""
        Returns a copy of this image whereby the mask is restricted to be equal
        to the convex hull around the chosen landmarks.

        The choice of whether a pixel is inside or outside of the pointcloud
        is determined by the ``point_in_pointcloud`` parameter. By default
        a Piecewise Affine transform is used to test for containment, which
        is useful when building efficiently aligning images. For large images,
        a faster and pixel-accurate method can be used ('convex_hull').
        Alternatively, a callable can be provided to override the test. By
        default, the provided implementations are only valid for 2D images.

        Parameters
        ----------
        group : `str`, optional
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.
            If the landmarks in question are an instance of :map:`TriMesh`,
            the triangulation of the landmarks will be used in the convex
            hull calculation. If the landmarks are an instance of
            :map:`PointCloud`, Delaunay triangulation will be used to
            create a triangulation.
        batch_size : `int` or ``None``, optional
            This should only be considered for large images. Setting this value
            will cause constraining to become much slower. This size indicates
            how many points in the image should be checked at a time, which
            keeps memory usage low. If ``None``, no batching is used and all
            points are checked at once. By default, this is only used for
            the 'pwa' point_in_pointcloud choice.
        point_in_pointcloud : {'pwa', 'convex_hull'} or `callable`
            The method used to check if pixels in the image fall inside the
            pointcloud or not. Can be accurate to a Piecewise Affine transform,
            a pixel accurate convex hull or any arbitrary callable.
            If a callable is passed, it should take two parameters,
            the :map:`PointCloud` to constrain with and the pixel locations
            ((d, n_dims) ndarray) to test and should return a (d, 1) boolean
            ndarray of whether the pixels were inside (True) or outside (False)
            of the :map:`PointCloud`.

        Returns
        -------
        constrained : :map:`MaskedImage`
            A new image where the mask is constrained by the provided
            landmarks.
        """
        copy = self.copy()
        copy.mask = copy.mask.constrain_to_pointcloud(
            copy.landmarks[group], batch_size=batch_size,
            point_in_pointcloud=point_in_pointcloud)
        return copy

    def build_mask_around_landmarks(self, patch_shape, group=None):
        r"""
        Deprecated - please use the equivalent
        `constrain_mask_to_patches_around_landmarks` method.
        """
        warn('This method is no longer supported and will be removed in a '
             'future version of Menpo. '
             'Use .constrain_mask_to_patches_around_landmarks() instead.',
             MenpoDeprecationWarning)
        return self.constrain_mask_to_patches_around_landmarks(
            patch_shape=patch_shape, group=group)

    def constrain_mask_to_patches_around_landmarks(self, patch_shape,
                                                   group=None):
        r"""
        Returns a copy of this image whereby the mask is restricted to be
        patches around each landmark in the chosen landmark group. The
        patch will be centred on the nearest pixel for each point in
        the chosen landmark group.

        Parameters
        ----------
        patch_shape : `tuple`
            The size of the patch.
        group : `str`, optional
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.

        Returns
        -------
        constrained : :map:`MaskedImage`
            A new image where the mask is constrained as patches centred on each
            point in the provided landmarks.
        """
        copy = self.copy()
        # get the selected pointcloud
        pc = copy.landmarks[group]
        # temporarily set all mask values to False
        copy.mask.pixels[:] = False
        # create a patches array of the correct size, full of True values
        patches = np.ones((pc.n_points, 1, 1, int(patch_shape[0]),
                           int(patch_shape[1])), dtype=np.bool)
        # set True patches around pointcloud centers
        copy.mask = copy.mask.set_patches(patches, pc)
        return copy

    def set_boundary_pixels(self, value=0.0, n_pixels=1):
        r"""
        Returns a copy of this :map:`MaskedImage` for which n pixels along
        the its mask boundary have been set to a particular value. This is
        useful in situations where there is absent data in the image which
        can cause, for example, erroneous computations of gradient or features.

        Parameters
        ----------
        value : `float` or (n_channels, 1) ndarray
        n_pixels : `int`, optional
            The number of pixels along the mask boundary that will be set to 0.

        Returns
        -------
        new_image : :map:`MaskedImage`
            The copy of the image for which the ``n`` pixels along its mask
            boundary have been set to a particular value.
        """
        copy = self.copy()
        global binary_erosion
        if binary_erosion is None:
            from scipy.ndimage import binary_erosion  # expensive
        # Erode the edge of the mask in by one pixel
        eroded_mask = binary_erosion(copy.mask.mask, iterations=n_pixels)

        # replace the eroded mask with the diff between the two
        # masks. This is only true in the region we want to nullify.
        np.logical_and(~eroded_mask, copy.mask.mask, out=eroded_mask)
        # set all the boundary pixels to a particular value
        copy.pixels[..., eroded_mask] = value
        return copy

    def erode(self, n_pixels=1):
        r"""
        Returns a copy of this :map:`MaskedImage` in which the mask has been
        shrunk by n pixels along its boundary.

        Parameters
        ----------
        n_pixels : `int`, optional
            The number of pixels by which we want to shrink the mask along
            its own boundary.

        Returns
        -------
        eroded_image : :map:`MaskedImage`
            The copy of the masked image in which the mask has been shrunk
            by n pixels along its boundary.
        """
        global binary_erosion
        if binary_erosion is None:
            from scipy.ndimage import binary_erosion  # expensive
        # Erode the edge of the mask in by one pixel
        eroded_mask = binary_erosion(self.mask.mask, iterations=n_pixels)

        image = self.copy()
        image.mask = BooleanImage(eroded_mask)
        return image

    def dilate(self, n_pixels=1):
        r"""
        Returns a copy of this :map:`MaskedImage` in which its mask has
        been expanded by n pixels along its boundary.

        Parameters
        ----------
        n_pixels : `int`, optional
            The number of pixels by which we want to expand the mask along
            its own boundary.

        Returns
        -------
        dilated_image : :map:`MaskedImage`
            The copy of the masked image in which the mask has been expanded
            by n pixels along its boundary.
        """
        global binary_dilation
        if binary_dilation is None:
            from scipy.ndimage import binary_dilation  # expensive
        # Erode the edge of the mask in by one pixel
        dilated_mask = binary_dilation(self.mask.mask, iterations=n_pixels)

        image = self.copy()
        image.mask = BooleanImage(dilated_mask)
        return image

    def rasterize_landmarks(self, group=None, render_lines=True, line_style='-',
                            line_colour='b', line_width=1, render_markers=True,
                            marker_style='o', marker_size=1,
                            marker_face_colour='b', marker_edge_colour='b',
                            marker_edge_width=1, backend='matplotlib'):
        r"""
        This method provides the ability to rasterize 2D landmarks onto the
        image. The returned image has the specified landmark groups rasterized
        onto the image - which is useful for things like creating result
        examples or rendering videos with annotations.

        Since multiple landmark groups can be specified, all arguments can take
        lists of parameters that map to the provided groups list. Therefore, the
        parameters must be lists of the correct length or a single parameter to
        apply to every landmark group.

        Multiple backends are provided, all with different strengths. The
        'pillow' backend is very fast, but not very flexible. The `matplotlib`
        backend should be feature compatible with other Menpo rendering methods,
        but is much slower due to the overhead of creating a figure to render
        into.

        Images will always be rendered masked with a black background.
        If an unmasked image is required, please use :meth:`as_unmasked`.

        Parameters
        ----------
        group : `str` or `list` of `str`, optional
            The landmark group key, or a list of keys.
        render_lines : `bool`, optional
            If ``True``, and the provided landmark group is a
            :map:`PointDirectedGraph`, the edges are rendered.
        line_style : `str`, optional
            The style of the edge line. Not all backends support this argument.
        line_colour : `str` or `tuple`, optional
            A Matplotlib style colour or a backend dependant colour.
        line_width : `int`, optional
            The width of the line to rasterize.
        render_markers : `bool`, optional
            If ``True``, render markers at the coordinates of each landmark.
        marker_style : `str`, optional
            A Matplotlib marker style. Not all backends support all marker
            styles.
        marker_size : `int`, optional
            The size of the marker - different backends use different scale
            spaces so consistent output may by difficult.
        marker_face_colour : `str`, optional
            A Matplotlib style colour or a backend dependant colour.
        marker_edge_colour : `str`, optional
            A Matplotlib style colour or a backend dependant colour.
        marker_edge_width : `int`, optional
            The width of the marker edge. Not all backends support this.
        backend : {'matplotlib', 'pillow'}, optional
            The backend to use.

        Returns
        -------
        rasterized_image : :map:`Image`
            The image with the landmarks rasterized directly into the pixels.

        Raises
        ------
        ValueError
            Only 2D images are supported.
        ValueError
            Only RGB (3-channel) or Greyscale (1-channel) images are supported.
        """
        from .rasterize import rasterize_landmarks_2d
        # Ensure that the image is ALWAYS masked - to make it consistent
        # between backends - the background will be black. as_unmasked should be
        # used to fiddle with the background colour.
        im = self.as_unmasked(copy=True, fill=0)
        return rasterize_landmarks_2d(
            im, group=group, render_lines=render_lines,
            line_style=line_style, line_colour=line_colour,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, backend=backend)



import numpy as np
from pytest import raises
from numpy.testing import (assert_allclose, assert_equal, assert_almost_equal,
                           assert_array_almost_equal)
from menpo.shape import PointCloud
from menpo.model import LinearVectorModel, PCAModel, PCAVectorModel
from menpo.math import as_matrix


def test_linear_model_creation():
    data = np.zeros((3, 120))
    LinearVectorModel(data)


def test_linear_model_basics():
    data = np.random.random((3, 120))
    linear_model = LinearVectorModel(data)
    assert(linear_model.n_components == 3)
    assert(linear_model.n_components == 3)
    assert(linear_model.n_features == 120)


def test_linear_model_project_vector():
    data = np.zeros((3, 120))
    data[0, 0] = 1
    data[1, 1] = 1
    data[2, 2] = 1
    linear_model = LinearVectorModel(data)
    sample = np.random.random(120)
    weights = linear_model.project(sample)
    assert_allclose(weights, sample[:3])


def test_linear_model_component():
    data = np.random.random((3, 120))
    linear_model = LinearVectorModel(data)
    assert_equal(linear_model.component(2), data[2])


def test_linear_model_instance_vector():
    data = np.zeros((3, 120))
    data[0, 0] = 1
    data[1, 1] = 1
    data[2, 2] = 1
    linear_model = LinearVectorModel(data)
    weights = np.array([0.263, 7, 41.2])
    projected = linear_model.instance(weights)
    # only the first 3 features are non zero...
    assert_allclose(projected[:3], weights)
    # rest should be nil
    assert_allclose(projected[3:], 0)


def test_pca_n_active_components():
    samples = [np.random.randn(10) for _ in range(10)]
    model = PCAVectorModel(samples)
    # integer
    model.n_active_components = 5
    assert_equal(model.n_active_components, 5)


def test_pca_n_active_components_too_many():
    samples = [np.random.randn(10) for _ in range(10)]
    model = PCAVectorModel(samples)
    # too many components
    model.n_active_components = 100
    assert_equal(model.n_active_components, 9)
    # reset too smaller number of components
    model.n_active_components = 5
    assert_equal(model.n_active_components, 5)
    # reset to too many components
    model.n_active_components = 100
    assert_equal(model.n_active_components, 9)


def test_pca_n_active_components_negative():
    samples = [np.random.randn(10) for _ in range(10)]
    model = PCAVectorModel(samples)
    with raises(ValueError):
        model.n_active_components = -5


def test_pca_trim():
    samples = [PointCloud(np.random.randn(10)) for _ in range(10)]
    model = PCAModel(samples)
    # trim components
    model.trim_components(5)
    # number of active components should be the same as number of components
    assert_equal(model.n_active_components, model.n_components)


def test_pca_trim_variance_limit():
    samples = [np.random.randn(10) for _ in range(10)]
    model = PCAVectorModel(samples)
    with raises(ValueError):
        # impossible to keep more than 1.0 ratio variance
        model.trim_components(2.5)


def test_pca_trim_variance_limit():
    samples = [np.random.randn(10) for _ in range(10)]
    model = PCAVectorModel(samples)
    with raises(ValueError):
        # impossible to keep more than 1.0 ratio variance
        model.trim_components(2.5)


def test_pca_trim_negative_integers():
    samples = [PointCloud(np.random.randn(10)) for _ in range(10)]
    model = PCAModel(samples)
    with raises(ValueError):
        # no negative number of components
        model.trim_components(-2)


def test_pca_trim_negative_float():
    samples = [PointCloud(np.random.randn(10)) for _ in range(10)]
    model = PCAModel(samples)
    with raises(ValueError):
        # no negative number of components
        model.trim_components(-2)


def test_pca_variance():
    samples = [PointCloud(np.random.randn(10)) for _ in range(10)]
    model = PCAModel(samples)
    # kept variance must be equal to total variance
    assert_equal(model.variance(), model.original_variance())
    # kept variance ratio must be 1.0
    assert_equal(model.variance_ratio(), 1.0)
    # noise variance must be 0.0
    assert_equal(model.noise_variance(), 0.0)
    # noise variance ratio must be also 0.0
    assert_equal(model.noise_variance_ratio(), 0.0)


def test_pca_inverse_noise_variance():
    samples = [PointCloud(np.random.randn(10)) for _ in range(10)]
    model = PCAModel(samples)
    with raises(ValueError):
        # inverse noise_variance it's not computable
        model.inverse_noise_variance()


def test_pca_variance_after_change_n_active_components():
    samples = [np.random.randn(10) for _ in range(10)]
    model = PCAVectorModel(samples)
    # set number of active components
    model.n_active_components = 5
    # kept variance must be smaller than total variance
    assert(model.variance() < model.original_variance())
    # kept variance ratio must be smaller than 1.0
    assert(model.variance_ratio() < 1.0)
    # noise variance must be bigger than 0.0
    assert(model.noise_variance() > 0.0)
    # noise variance ratio must also be bigger than 0.0
    assert(model.noise_variance_ratio() > 0.0)
    # inverse noise variance is computable
    assert(model.inverse_noise_variance() == 1/model.noise_variance())


def test_pca_variance_after_trim():
    samples = [np.random.randn(10) for _ in range(10)]
    model = PCAVectorModel(samples)
    # set number of active components
    model.trim_components(5)
    # kept variance must be smaller than total variance
    assert(model.variance() < model.original_variance())
    # kept variance ratio must be smaller than 1.0
    assert(model.variance_ratio() < 1.0)
    # noise variance must be bigger than 0.0
    assert(model.noise_variance() > 0.0)
    # noise variance ratio must also be bigger than 0.0
    assert(model.noise_variance_ratio() > 0.0)
    # inverse noise variance is computable
    assert(model.inverse_noise_variance() == 1 / model.noise_variance())


def test_pca_orthogonalize_against():
    pca_samples = np.random.randn(10, 10)
    pca_model = PCAVectorModel(pca_samples)
    lm_samples = np.asarray([np.random.randn(10) for _ in range(4)])
    lm_model = LinearVectorModel(np.asarray(lm_samples))
    # orthogonalize
    pca_model.orthonormalize_against_inplace(lm_model)
    # number of active components must remain the same
    assert_equal(pca_model.n_active_components, 6)


def test_pca_orthogonalize_against_with_less_active_components():
    pca_samples = np.random.randn(10, 10)
    pca_model = PCAVectorModel(pca_samples)
    lm_samples = np.asarray([np.random.randn(10) for _ in range(4)])
    lm_model = LinearVectorModel(np.asarray(lm_samples))
    # set number of active components
    pca_model.n_active_components = 5
    # orthogonalize
    pca_model.orthonormalize_against_inplace(lm_model)
    # number of active components must remain the same
    assert_equal(pca_model.n_active_components, 5)


def test_pca_increment_centred():
    pca_samples = [PointCloud(np.random.randn(10, 2)) for _ in range(10)]
    ipca_model = PCAModel(pca_samples[:3])
    ipca_model.increment(pca_samples[3:6])
    ipca_model.increment(pca_samples[6:])

    bpca_model = PCAModel(pca_samples)

    assert_almost_equal(np.abs(ipca_model.components),
                        np.abs(bpca_model.components))
    assert_almost_equal(ipca_model.eigenvalues, bpca_model.eigenvalues)
    assert_almost_equal(ipca_model.mean().as_vector(),
                        bpca_model.mean().as_vector())


def test_pca_increment_noncentred():
    pca_samples = [np.random.randn(10) for _ in range(10)]
    ipca_model = PCAVectorModel(pca_samples[:3], centre=False)
    ipca_model.increment(pca_samples[3:6])
    ipca_model.increment(pca_samples[6:])

    bpca_model = PCAVectorModel(pca_samples, centre=False)

    assert_almost_equal(np.abs(ipca_model.components),
                        np.abs(bpca_model.components))
    assert_almost_equal(ipca_model.eigenvalues, bpca_model.eigenvalues)
    assert_almost_equal(ipca_model.mean(), bpca_model.mean())


def test_pca_vector_init_from_covariance():
    n_samples = 30
    n_features = 10
    centre_values = [True, False]
    for centre in centre_values:
        # generate samples matrix and mean vector
        samples = np.random.randn(n_samples, n_features)
        mean = np.mean(samples, axis=0)
        # compute covariance matrix
        if centre:
            X = samples - mean
            C = np.dot(X.T, X) / (n_samples - 1)
        else:
            C = np.dot(samples.T, samples) / (n_samples - 1)
        # create the 2 pca models
        pca1 = PCAVectorModel.init_from_covariance_matrix(C, mean, centred=centre,
                                                          n_samples=n_samples)
        pca2 = PCAVectorModel(samples, centre=centre, inplace=False)
        # compare them
        assert_array_almost_equal(pca1.mean(), pca2.mean())
        assert_array_almost_equal(pca1.component(0, with_mean=False),
                                  pca2.component(0, with_mean=False))
        assert_array_almost_equal(pca1.component(7), pca2.component(7))
        assert_array_almost_equal(pca1.components, pca2.components)
        assert_array_almost_equal(pca1.eigenvalues, pca2.eigenvalues)
        assert_array_almost_equal(pca1.eigenvalues_cumulative_ratio(),
                                  pca2.eigenvalues_cumulative_ratio())
        assert_array_almost_equal(pca1.eigenvalues_ratio(),
                                  pca2.eigenvalues_ratio())
        weights = np.random.randn(pca1.n_active_components - 4)
        assert_array_almost_equal(pca1.instance(weights),
                                  pca2.instance(weights))
        assert(pca1.n_active_components == pca2.n_active_components)
        assert(pca1.n_components == pca2.n_components)
        assert(pca1.n_features == pca2.n_features)
        assert(pca1.n_samples == pca2.n_samples)
        assert(pca1.noise_variance() == pca2.noise_variance())
        assert(pca1.noise_variance_ratio() == pca2.noise_variance_ratio())
        assert_allclose(pca1.variance(), pca2.variance())
        assert(pca1.variance_ratio() == pca2.variance_ratio())
        assert_array_almost_equal(pca1.whitened_components(),
                                  pca2.whitened_components())


def test_pca_init_from_covariance():
    n_samples = 30
    n_features = 10
    n_dims = 2
    centre_values = [True, False]
    for centre in centre_values:
        # generate samples list and convert it to nd.array
        samples = [PointCloud(np.random.randn(n_features, n_dims))
                   for _ in range(n_samples)]
        data, template = as_matrix(samples, return_template=True)
        # compute covariance matrix and mean
        if centre:
            mean_vector = np.mean(data, axis=0)
            mean = template.from_vector(mean_vector)
            X = data - mean_vector
            C = np.dot(X.T, X) / (n_samples - 1)
        else:
            mean = samples[0]
            C = np.dot(data.T, data) / (n_samples - 1)
        # create the 2 pca models
        pca1 = PCAModel.init_from_covariance_matrix(C, mean,
                                                    centred=centre,
                                                    n_samples=n_samples)
        pca2 = PCAModel(samples, centre=centre)
        # compare them
        assert_array_almost_equal(pca1.component_vector(0, with_mean=False),
                                  pca2.component_vector(0, with_mean=False))
        assert_array_almost_equal(pca1.component(7).as_vector(),
                                  pca2.component(7).as_vector())
        assert_array_almost_equal(pca1.components, pca2.components)
        assert_array_almost_equal(pca1.eigenvalues, pca2.eigenvalues)
        assert_array_almost_equal(pca1.eigenvalues_cumulative_ratio(),
                                  pca2.eigenvalues_cumulative_ratio())
        assert_array_almost_equal(pca1.eigenvalues_ratio(),
                                  pca2.eigenvalues_ratio())
        weights = np.random.randn(pca1.n_active_components)
        assert_array_almost_equal(pca1.instance(weights).as_vector(),
                                  pca2.instance(weights).as_vector())
        weights2 = np.random.randn(pca1.n_active_components - 4)
        assert_array_almost_equal(pca1.instance_vector(weights2),
                                  pca2.instance_vector(weights2))
        assert_array_almost_equal(pca1.mean().as_vector(),
                                  pca2.mean().as_vector())
        assert_array_almost_equal(pca1.mean_vector,
                                  pca2.mean_vector)
        assert(pca1.n_active_components == pca2.n_active_components)
        assert(pca1.n_components == pca2.n_components)
        assert(pca1.n_features == pca2.n_features)
        assert(pca1.n_samples == pca2.n_samples)
        assert(pca1.noise_variance() == pca2.noise_variance())
        assert(pca1.noise_variance_ratio() == pca2.noise_variance_ratio())
        assert_almost_equal(pca1.variance(), pca2.variance())
        assert_almost_equal(pca1.variance_ratio(), pca2.variance_ratio())
        assert_array_almost_equal(pca1.whitened_components(),
                                  pca2.whitened_components())


def test_pca_project():
    pca_samples = [PointCloud(np.random.randn(10, 2)) for _ in range(10)]
    pca_model = PCAModel(pca_samples)
    projected = pca_model.project(pca_samples[0])
    assert projected.shape[0] == 9



from .input import (
    import_image, import_images, image_paths,
    import_video, import_videos, video_paths,
    import_landmark_file, import_landmark_files, landmark_file_paths,
    import_pickle, import_pickles, pickle_paths,
    import_builtin_asset, data_dir_path, data_path_to, ls_builtin_assets,
    register_image_importer, register_landmark_importer,
    register_pickle_importer, register_video_importer
)
from .output import (export_image, export_video,
                     export_landmark_file, export_pickle)
from .exceptions import OverwriteError



import numpy as np

map_coordinates = None  # expensive, from scipy.ndimage
from menpo.transform import Homogeneous

# Store out a transform that simply switches the x and y axis
xy_yx = Homogeneous(np.array([[0., 1., 0.],
                              [1., 0., 0.],
                              [0., 0., 1.]]))


def scipy_interpolation(pixels, points_to_sample, mode='constant', order=1,
                        cval=0.):
    r"""
    Interpolation utilizing scipy map_coordinates function.

    Parameters
    ----------
    pixels : ``(n_channels, M, N, ...)`` `ndarray`
        The image to be sampled from, the first axis containing channel
        information
    points_to_sample : ``(n_points, n_dims)`` `ndarray`
        The points which should be sampled from pixels
    mode : ``{constant, nearest, reflect, wrap}``, optional
        Points outside the boundaries of the input are filled according to the
        given mode
    order : `int,` optional
        The order of the spline interpolation. The order has to be in the
        range [0, 5].
    cval : `float`, optional
        The value that should be used for points that are sampled from
        outside the image bounds if mode is ``constant``.

    Returns
    -------
    sampled_image : `ndarray`
        The pixel information sampled at each of the points.
    """
    global map_coordinates
    if map_coordinates is None:
        from scipy.ndimage import map_coordinates  # expensive
    sampled_pixel_values = np.empty((pixels.shape[0], points_to_sample.shape[0]),
                                    dtype=pixels.dtype)

    # Loop over every channel in image - we know first axis is always channels
    # Note that map_coordinates uses the opposite (dims, points) convention
    # to us so we transpose
    points_to_sample_t = points_to_sample.T
    for i in range(pixels.shape[0]):
        map_coordinates(pixels[i],
                        points_to_sample_t,
                        mode=mode,
                        order=order,
                        cval=cval,
                        output=sampled_pixel_values[i])
    return sampled_pixel_values


try:
    import cv2


    def _mode_to_opencv(mode):
        if mode == 'nearest':
            return cv2.BORDER_REPLICATE
        elif mode == 'constant':
            return cv2.BORDER_CONSTANT
        else:
            raise ValueError('Unknown mode "{}", must be one of '
                             '(nearest, constant)'.format(mode))


    def _order_to_opencv(order):
        if order == 0:
            return cv2.INTER_NEAREST
        elif order == 1:
            return cv2.INTER_LINEAR
        else:
            raise ValueError('Unsupported order "{}", '
                             'must be one of (0, 1)'.format(order))


    def cv2_perspective_interpolation(pixels, template_shape, h_transform,
                                      mode='constant', order=1, cval=0.):
        r"""
        Interpolation utilizing OpenCV fast perspective warping. This method
        assumes that the warp takes the form of a homogeneous transform, and
        thus is much faster for operations such as scaling.

        Note less modes and orders are supported than the more generic
        interpolation methods.

        Parameters
        ----------
        pixels : ``(n_channels, M, N, ...)`` `ndarray`
            The image to be sampled from, the first axis containing channel
            information.
        template_shape : `tuple`
            The shape of the new image that will be sampled
        mode : ``{constant, nearest}``, optional
            Points outside the boundaries of the input are filled according to the
            given mode.
        order : int, optional
            The order of the spline interpolation. The order has to be in the
            range [0, 1].
        cval : `float`, optional
            The value that should be used for points that are sampled from
            outside the image bounds if mode is 'constant'

        Returns
        -------
        sampled_image : `ndarray`
            The pixel information sampled at each of the points.
        """
        matrix = xy_yx.compose_before(h_transform).compose_before(xy_yx).h_matrix
        # Ensure template shape is a tuple (as required by OpenCV)
        template_shape = tuple(template_shape)
        cv_template_shape = template_shape[::-1]  # Flip to (W, H)

        # Unfortunately, OpenCV does not seem to support the boolean numpy type
        if pixels.dtype == np.bool:
            in_pixels = pixels.astype(np.uint8)
        else:
            in_pixels = pixels

        warped_image = np.empty((pixels.shape[0],) + template_shape,
                                dtype=in_pixels.dtype)
        for i in range(pixels.shape[0]):
            cv2.warpPerspective(in_pixels[i], matrix, cv_template_shape,
                                dst=warped_image[i],
                                flags=_order_to_opencv(order) + cv2.WARP_INVERSE_MAP,
                                borderMode=_mode_to_opencv(mode),
                                borderValue=cval)

        # As above, we may need to convert the uint8 back to bool
        if pixels.dtype == np.bool:
            warped_image = warped_image.astype(np.bool)
        return warped_image
except ImportError:
    pass



import numpy as np
from menpo.shape import TriMesh


def utils_mesh():
    trilist = np.array([[0, 1, 2],
                        [0, 2, 3]])
    points = np.array([[0, 0],
                       [1, 0],
                       [1, 1],
                       [0, 1]])
    return TriMesh(points, trilist=trilist)


gt_edge_indices = np.array([[0, 1],
                            [1, 2],
                            [2, 0],
                            [0, 2],
                            [2, 3],
                            [3, 0]])

gt_edge_vectors = np.array([[1, 0],
                            [0, 1],
                            [1, 1],
                            [1, 1],
                            [-1, 0],
                            [0, 1]])

gt_unique_edge_vectors = np.array([[1, 0],
                                   [1, 1],
                                   [0, 1],
                                   [0, 1],
                                   [-1, 0]])

gt_edge_lengths = np.array([1., 1., 1.41421356, 1.41421356, 1., 1.])

gt_unique_edge_lengths = np.array([1., 1.41421356, 1., 1., 1.])

gt_tri_areas = np.array([0.5, 0.5])


def test_edge_indices():
    np.testing.assert_allclose(utils_mesh().edge_indices(), gt_edge_indices)


def test_edge_vectors():
    np.testing.assert_allclose(utils_mesh().edge_vectors(), gt_edge_vectors)


def test_unique_edge_vectors():
    np.testing.assert_allclose(utils_mesh().unique_edge_vectors(),
                               gt_unique_edge_vectors)


def test_edge_lengths():
    assert np.allclose(utils_mesh().edge_lengths(), gt_edge_lengths)


def test_unique_edge_lengths():
    np.testing.assert_allclose(utils_mesh().unique_edge_lengths(),
                               gt_unique_edge_lengths)


def test_tri_areas():
    np.testing.assert_allclose(utils_mesh().tri_areas(), gt_tri_areas)


def test_2d_trimesh_2d_positive_areas():
    t = TriMesh(np.array([[0, 0], [0, 1],
                          [1, 1], [1, 0]], dtype=np.float),
                trilist=np.array([[0, 2, 3], [0, 2, 1]]))
    assert np.all(t.tri_areas() > 0)


def test_mean_tri_area():
    assert np.isclose(utils_mesh().mean_tri_area(), 0.5)


def test_mean_edge_length():
    np.testing.assert_allclose(utils_mesh().mean_edge_length(),
                               np.mean(gt_unique_edge_lengths))


def test_mean_edge_length_not_unique():
    np.testing.assert_allclose(utils_mesh().mean_edge_length(unique=False),
                               np.mean(gt_edge_lengths))



import numpy as np
from scipy.spatial.distance import cdist
from .base import Transform


class RadialBasisFunction(Transform):
    r"""
    Radial Basis Functions are a class of transform that is used by
    :map:`ThinPlateSplines`. They have to be able to take their own radial
    derivative for :map:`ThinPlateSplines` to be able to take its own total
    derivative.

    Parameters
    ----------
    c : ``(n_centres, n_dims)`` `ndarray`
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """
    def __init__(self, c):
        self.c = c

    @property
    def n_centres(self):
        r"""
        The number of centres.

        :type: `int`
        """
        return self.c.shape[0]

    @property
    def n_dims(self):
        r"""
        The RBF can only be applied on points with the same dimensionality as
        the centres.

        :type: `int`
        """
        return self.c.shape[1]

    @property
    def n_dims_output(self):
        r"""
        The result of the transform has a dimension (weight) for every centre.

        :type: `int`
        """
        return self.n_centres


class R2LogR2RBF(RadialBasisFunction):
    r"""
    The :math:`r^2 \log{r^2}` basis function.

    The derivative of this function is :math:`2 r (\log{r^2} + 1)`.

    .. note::

        :math:`r = \lVert x - c \rVert`

    Parameters
    ----------
    c : ``(n_centres, n_dims)`` `ndarray`
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """
    def __init__(self, c):
        super(R2LogR2RBF, self).__init__(c)

    def _apply(self, x, **kwargs):
        """
        Apply the basis function.

        .. note::

            :math:`r^2 \log{r^2} === r^2 2 \log{r}`

        Parameters
        ----------
        x : ``(n_points, n_dims)`` `ndarray`
            Set of points to apply the basis to.

        Returns
        -------
        u : ``(n_points, n_centres)`` `ndarray`
            The basis function applied to each distance,
            :math:`\lVert x - c \rVert`.
        """
        euclidean_distance = cdist(x, self.c)
        mask = euclidean_distance == 0
        with np.errstate(divide='ignore', invalid='ignore'):
            u = (euclidean_distance ** 2 *
                 (2 * np.log(euclidean_distance)))
        # reset singularities to 0
        u[mask] = 0
        return u


class R2LogRRBF(RadialBasisFunction):
    r"""
    Calculates the :math:`r^2 \log{r}` basis function.

    The derivative of this function is :math:`r (1 + 2 \log{r})`.

    .. note::

        :math:`r = \lVert x - c \rVert`

    Parameters
    ----------
    c : ``(n_centres, n_dims)`` `ndarray`
        The set of centers that make the basis. Usually represents a set of
        source landmarks.
    """
    def __init__(self, c):
        super(R2LogRRBF, self).__init__(c)

    def _apply(self, points, **kwargs):
        """
        Apply the basis function :math:`r^2 \log{r}`.

        Parameters
        ----------
        points : ``(n_points, n_dims)`` `ndarray`
            Set of points to apply the basis to.

        Returns
        -------
        u : ``(n_points, n_centres)`` `ndarray`
            The basis function applied to each distance,
            :math:`\lVert points - c \rVert`.
        """
        euclidean_distance = cdist(points, self.c)
        mask = euclidean_distance == 0
        with np.errstate(divide='ignore', invalid='ignore'):
            u = euclidean_distance ** 2 * np.log(euclidean_distance)
        # reset singularities to 0
        u[mask] = 0
        return u
